/*
 * initialise a risc-v RV64G system enough to call main with the low identity
 * map and upper addresses mapped to lower, in supervisor mode.
 * this is traditionally done entirely in assembler, typically called l.s.
 *
 * do not use print nor iprint with '%' in the format here; libc/fmt learns
 * function addresses for verbs, which need to be high addresses.
 *
 * all cpus may be executing this code simultaneously.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"

#include "riscv64.h"

#define dbprf if (!Lowdebug) {} else prf

#ifndef Lowdebug
#define Lowdebug 0	/* beware: non-0 may throw off MP start-up timing */
#endif

enum {
	Diagnose = 0,		/* self-checking for a new machine */
	Measurestkuse = 0,

	Pmpdebug = 0,
	Configpmp = 0,		/* don't need this so far */
	Blindpmpsetting = 0,

	Pmpwid = 8,		/* bits per slot in cfg reg */
	Pmpslots = 8,		/* slots per cfg reg */
	Pmpmask = VMASK(Pmpwid),

	Ptedebug = 0,		/* flag: print initial PTEs */
	Nptes = 12,		/* number of Ptes to dump */
	// Nptes = Ptpgptes / 2,

	Pageexcs = 1<<Instpage | 1<<Loadpage | 1<<Storepage,
};

/* init with anything to force into data seg. to avoid bss zeroing */
int	asids = 0;
int	early = 1;			/* tested in trap.c */
ushort	hartids[MACHMAX] = { 0 };
Rvarch *initarchp;
char	initstks[MACHMAX][INITSTKSIZE] = { 0 };
uintptr	mainpc = 0;
uintptr satptval, satpepc;
void	(*tlbinvall)(void);

static	int npmpcfg;

static void mkinitpgtbl(Sys *lowsys);

/*
 * machine mode setup and switch to supervisor mode.
 */

/* c910 needs us to write PMPCFG regs before we can read actual contents. */
static uvlong
rdpmpcfg(int n)
{
	m->probebad = 0;
	csrset(PMPCFG + n, 0);
	return csrrd(PMPCFG + n);
}

static uvlong
rdpmpslot(uint slot)
{
	uvlong cfg;

	cfg = rdpmpcfg(2*(slot/Pmpslots));
	return (cfg >> ((slot%Pmpslots)*Pmpwid)) & Pmpmask;
}

/* work out number of pmp config csrs */
static int
getnpmpcfgs(void)
{
	rdpmpcfg(0);			/* see if it faults */
	if (m->probebad) {
		m->probebad = 0;
		return 0;
	}
	if (npmpcfg == 0) {
		rdpmpcfg(2);		/* see if it faults */
		if (m->probebad) {
			m->probebad = 0;
			return 2;
		} else
			return 8;
	}
	return npmpcfg;
}

static void
dumppmp(char *desc)
{
	int slot;
	uint reg;
	uvlong cfg, addr;

	USED(desc);
	if (!Pmpdebug)
		return;
	for (reg = 0; reg < npmpcfg; reg++) {
		cfg = rdpmpcfg(2*reg);
		if (cfg == 0)
			continue;
		dbprf("\n%s: pmpcfg[%d] %#p\n", desc, 2*reg, cfg);
		for (slot = 0; slot < Pmpslots; slot++) {
			addr = csrrd(PMPADDR+Pmpslots*reg+slot);
			if (addr)
				dbprf("pmpaddr%d %#p\n", slot, addr);
		}
	}
	USED(cfg);
}

/*
 * pmp is optional, and somewhat redundant on systems with MMUs.
 * it's likely to be configured correctly already since u-boot or other
 * boot loader has normally run.
 *
 * make sure that the whole of ram is rwx, where possible.  pmp csrs are only
 * accessible in machine mode, and pmp grants access only to accesses in S and
 * U modes, but pmp can restrict accesses in M mode.
 *
 * there may be 0, 16 or 64 entries, and only even-numbered PMPCFG registers
 * are used in RV64.
 */
static void
pmpinit(void)
{
	uint reg, slot, slotsh;

	/* work out number of pmp config csrs */
	dbprf("pmpinit...");
	npmpcfg = getnpmpcfgs();
	if (npmpcfg == 0) {
		dbprf("no PMP...");
		return;
	}

	dumppmp("initial");
	if (!Configpmp)
		return;

	wbinvd();
	if (Blindpmpsetting) {			/* cheap hack */
		csrswap(PMPADDR+0, ~0ULL);
		csrswap(PMPCFG+0, Pmpnapot|Pmpr|Pmpw|Pmpx);
	} else {
		/* find first free slot, if any */
		for (slot = 0; slot < npmpcfg*Pmpslots; slot++)
			if (rdpmpslot(slot) == 0)
				break;
		if (slot >= npmpcfg*Pmpslots) {
			dbprf("no free pmpcfg slot in %d regs\n", npmpcfg);
			return;
		}

		/*
		 * if next slot is free too, leave this one zero as base addr.
		 * special case: Pmptor in slot 0 uses zero base.
		 */
		if (slot > 0 && slot < npmpcfg*Pmpslots-1 &&
		    rdpmpslot(slot+1) == 0)
			slot++;
		dbprf("free pmpcfg slot %d; allowing all\n", slot);

		/* 56 bits max. phys., low 2 implied */
		csrswap(PMPADDR + slot, VMASK(56-2));
		reg = 2 * (slot / Pmpslots);
		slotsh = (slot % Pmpslots) * Pmpwid;
		/*
		 * if we add Pmplock to cfg, this setting will apply even to M
		 * mode, but can't be changed until reset or power cycle.
		 */
		csrset(PMPCFG + reg, (Pmptor|Pmpr|Pmpw|Pmpx)<<slotsh);
	}
	wbinvd();		/* sfence.vma is required since priv 1.12 */
	dumppmp("configured");
}

/* write csr then read back its value to see what stuck */
uintptr
csrwrrd(ushort csrno, uintptr new)
{
	csrswap(csrno, new);
	return csrrd(csrno);
}

/*
 * delegate M-mode exceptions and interrupts to S-mode, when possible,
 * and allow access to cycle counters.  also configure PMP if present.
 */
static void
delegate(void)
{
	dbprf("deleg...");
	csrswap(MCOUNTEREN, ~0ull); /* expose timers & cycle counters to S */
	csrswap(SCOUNTEREN, ~0ull);
//	mideleg = csrwrrd(MIDELEG, Superie);	/* old: punt S intrs to S */
	mideleg = csrwrrd(MIDELEG, ~0ull); /* try to punt S+M intrs to S */

	/* catch early stray M faults */
	dbprf("mtraps abort...");
	putmtvec(recmtrapalign);		/* low due to low PC */

	/* don't punt any except.s so we can detect PMP presence */
	csrswap(MEDELEG, 0);
	pmpinit();

	/* punt except.s to S except env call from S (for reboot) */
	medeleg = csrwrrd(MEDELEG, ~(1LL<<Envcallsup));
	if ((medeleg & Pageexcs) != Pageexcs)
		prf("page fault exceptions not delegated by M mode!\n");
}

#define EXT(let) (1LL << ((let) - 'A'))

static void
loadids(Sys *lowsys)
{
	lowsys->extensions = csrrd(MISA);
	lowsys->archid =   csrrd(MARCHID);
	lowsys->vendorid = csrrd(MVENDORID);
	dbprf("misa %#p marchid %#p mvendorid %#lux...",
		lowsys->extensions, lowsys->archid, lowsys->vendorid);

	if (lowsys->vendorid == Vthead)
		dbprf("xuantie cpu...");
	if (lowsys->extensions & EXT('H')) {
		dbprf("hypervisor enabled...");
		csrswap(MISA, lowsys->extensions & ~EXT('H'));
		dbprf("now disabled...");
	}
}

static int
readwd(ulong *ptr, ulong exp)
{
	ulong val;

	val = readmprv(ptr);
	if (val == exp) {
		dbprf("\nread %#p: got expected %#lux as Super...", ptr, val);
		return 0;
	} else {
		dbprf("\nread %#p: %#lux, expected %#lux as Super...",
			ptr, val, exp);
		return -1;
	}
}

static int
writewd(ulong *ptr, ulong val)
{
	writelmprv(ptr, val);
	return readwd(ptr, val);
}

static void
writelong(ulong *ptr, ulong val)
{
	dbprf(" writing %#p with %#lux as Super: ", ptr, val);
	delay(100);
	writelmprv(ptr, val);
	delay(100);
}

static void
satpprobe(Sys *lowsys)			/* only safe in machine mode */
{
	long r;
	uvlong mode, bestmode, readmode;
	static ulong testwd = -1;

	dbprf("\n");
	mkinitpgtbl(lowsys);

	/* probe satp modes */
	dbprf("probing satp modes...");
	bestmode = Sv39;
	asids = 0;
	for (mode = Sv39; mode <= Sv64; mode += 1LL<<Svshft) {
		putsatp(mode | Satpasidmask);
		readmode = getsatp();
		if ((readmode & ~Satpasidmask) == mode)
			bestmode = readmode;
	}
	asids = ((bestmode & Satpasidmask) >> Satpasidshft) + 1;
	dbprf("best available satp mode is %#llud with %d asids...",
		bestmode>>Svshft, asids);
	USED(bestmode);

	/* set up mmu, verify operation */
	dbprf("\nprobe initial page table (satp %#p)...", lowsys->satp);
	putsatp(lowsys->satp);
	if (getsatp() != lowsys->satp) {
		dbprf("can't set satp to %#p!\n", lowsys->satp);
		return;
	}
	dbprf("installed...");

	/*
	 * Mpp is Mppsuper, inherited from start.s.
	 */
	dbprf("\ntest mprv...");
	wbinvd();
	/* _main is low here, because PC is low. */
	/* load as if in S mode */
	enum {
		Textstart = 0x0ff0000f10017073LL, /* SPLHI: CSRRC, FENCE */
	};
	r  = readwd((ulong *)_main, Textstart);
	r |= readwd(ensurehigh(_main), Textstart);

	/*
	 * WTF?  uart writes do nothing on c910.  is i/o broken in S mode?
	 * works if we omit beginxt.c, mchxt.s & -DC910.  maybe we have to
	 * enable a zillion clock signals and take a zillion SoC components
	 * out of reset?
	 */
	writelong((ulong *)PAUart0, '*');
	writelong(ensurehigh(PAUart0), '~');

	/* writes via high addresses to normal memory work */
	writewd(ensurehigh(&testwd), 0x1234);

	dbprf("\npage table seems %s\n", r == 0? "okay": "bad");
	USED(r);
	lowsys->satp = 0;
	putsatp(0);
}

static void
assertrv(char *claim)
{
	USED(claim);
	dbprf("** If cpu is actually risc-v, %s.\n", claim);
}

static void
macharchinit(void)
{
	if (initarchp == nil)
		return;
	if (initarchp->machexten) {
		dbprf("\nenable M mode extensions...");
		(*(Pvfnv)ensurelow(initarchp->machexten))();
	}
	/* start other harts while we can */
	if (m->machno == 0 && initarchp->machstharts) {
		dbprf("start harts via M mode extensions...");
		(*(Pvfnv)ensurelow(initarchp->machstharts))();
	}
}

/*
 * do machine-mode set up, then leave machine mode.
 * set up for supervisor mode, needed for paging in kernel.
 * mostly punting faults to supervisor mode, preparing for clock interrupts,
 * and disabling paging.  mret will switch us to supervisor mode.
 *
 * needs m set, for CSR(MSCRATCH).
 */
static void
mach_to_super(Sys *lowsys)
{
	dbprf(" * machine mode.\n");
	delegate();
	/* see setstkmach0 for # of priv modes (lowsys->nprivmodes) */
	/* M mode: little endian, prev mode = super, allow super interrupts */
	putmsts(Defmsts & ~(Mie|Upie|Uie|Tvm|Tw|Tsr|Ube|Mbe64|Sbe64));
	if (getmsts() & Uie)
		prf("user interrupts enabled in M; can't be disabled\n");
	dbprf("mstatus %#p...", getmsts());
	loadids(lowsys);
	macharchinit();

	/* set up machine traps on this cpu.  Mach is soon to be populated. */
	dbprf("\nmtraps...");
	csrswap(MSCRATCH, (uintptr)m);	/* mmu will be off; low m & pc */
	/*
	 * maybe sbi is present, despite starting us in M mode?
	 * mtvec may have been from sbi, u-boot or rebootcode.
	 */
	dbprf("original mtvec %#p...", origmtvec);
	if (soc.c910 && origmtvec)
		putmtvec(origmtvec);	/* in case sbi is present */
	else
		putmtvec(mtrap);
	dbprf("mtvec to %#p\n", getmtvec());
	putstvec(rectrapalign);

	if (Diagnose && m->machno == 0)
		satpprobe(lowsys); /* find best mode & validate page table */

	/* switch to super mode & enable mach intrs. */
	splhi();
	putmie(getmie() | Machie);	/* needed for clock intrs */
	putmsts(getmsts() | Mie);  /* Mpp is still Mppsuper, from start.s */
	dbprf("to super: mstatus %#p sstatus %#p mret\n", getmsts(), getsts());
	assertrv("switching to supervisor mode");
	mret();
}

/*
 * the rest is supervisor mode setup.
 */

/*
 * populate n page table entries at ptep with attributes from ptebits
 * and physical addresses starting from ptebits.
 * lvl is the level of the page table.
 */
static void
setptes(PTE *ptep, PTE ptebits, int n, int lvl)
{
	PTE pteincr;

	pteincr = (PGLSZ(lvl)>>PGSHFT) << PTESHFT;
	for (; n > 0; n--) {
		*ptep++ = ptebits;
		ptebits += pteincr;
	}
}

/*
 * create top-level leaf page table entries in page table at ptp
 * to phys for phys & KZERO|phys.
 */
static void
dualmap(PTE *ptp, uintptr phys, uint nptes)
{
	PTE ptebits;

	/* force to physical space, round down to nearest superpage */
	phys = ROUNDDN(phys & ~KZERO, PGLSZ(Toplvl));
	ptebits = (phys>>PGSHFT)<<PTESHFT | PteR | PteW | PteX | Pteleafvalid;
	setptes(&ptp[PTLX(phys, Toplvl)], ptebits, nptes, Toplvl);
	setptes(&ptp[PTLX(KZERO+phys, Toplvl)], ptebits | PteG, nptes, Toplvl);
	coherence();
}

static void
dumpptes(PTE *ptp)
{
	int i, st;

	dbprf("initial page table excerpts:\n");
	for (i = 0; i < Nptes; i++)
		if (ptp[i])
			dbprf("ptp[%d] %#p\n", i, ptp[i]);
	st = PTLX(KZERO, Toplvl);
	for (i = st; i < st + Nptes; i++)
		if (ptp[i])
			dbprf("ptp[%d] %#p\n", i, ptp[i]);
}

#ifdef FILL_PT
#include "rvb/c910.fill.ptes.c"
#endif

/*
 * construct minimal initial top-level page table with id & upper->lower maps.
 */
static void
mkinitpgtbl(Sys *lowsys)
{
	uint nptes;
	PTE *ptp;

	if (lowsys->satp & pagingmode)		/* already made? */
		return;
	/*
	 * allocate temporary top-level page table.
	 * in top-level pt for 4 levels, each PTE covers 256GB.
	 * for 3 levels, each covers 1GB.
	 */
	ptp = lowsys->initpt;
	dbprf("creating initial page table at %#p\n", ptp);
	memset(lowsys->initpt, 0, sizeof lowsys->initpt);

	/*
	 * leave last pte of upper and lower ranges free
	 * (for VMAP in upper->lower map at least).
	 */
	nptes = Ptpgptes/2 - 1;		/* for each range */
	USED(nptes);
	if (VMBITS < 39)
		nptes /= 2;	/* upper and lower ranges fit into low addr.s */

	/* populate id map in lower range & upper->lower, from 0 up. */
	dualmap(ptp, 0, nptes);
#ifdef FILL_PT
	if (soc.c910)
		/* don't map first 4K or 2MB page to avoid hang */
		expandpte(lowsys, ptp, PteR|PteW|PteX | Pteleafvalid,
			PHYSMEM, 0);	// PPN((uvlong)KTZERO & VMASK(30)));
#endif
	if (Ptedebug)
		dumpptes(ptp);
	lowsys->satp = pagingmode | ((uintptr)ptp / PGSZ);
}

static void
setmtimecmp(int cpu)
{
	Clint *clint;

	/*
	 * assumes a standard clint.  m->timecmp is mainly for mtrap.s,
	 * use low addr.
	 */
	clint = m->clint;
	if (clint == nil) {
		m->mtimecmp = nil;
		prf("cpu%d: zero clint address\n", cpu);
	} else {
		/* too soon to probe for clint; this breaks on vf2 */
		if (0 && probeulong((ulong *)clint, Read) < 0)
			/* we won't get far without a working clint */
			prf("cpu%d: no response from clint at %#p\n",
				cpu, clint);
		else
			dbprf("cpu%d: clint at %#p\n", cpu, clint);
		m->mtimecmp = &clint->mtimecmp[m->hartid];
	}
}

static void
newmach(Mach *mp, uint cpu)
{
	/* initialise enough of Mach to get to main */
	m = mp;				/* low address */
	m->machmode = bootmachmode;
	fakecpuhz();
	usephysdevaddrs();
	m->machno = cpu;
	m->hartid = hartids[cpu];

	if (!bootmachmode)
		putsscratch((uintptr)m);	/* ready for straps now */
	setmtimecmp(cpu);
	putsscratch((uintptr)m);		/* ready for straps now */
	if(cpu == 0)
		dbprf("Booting Plan 9 on hart %d\n", m->hartid);
	clrstie();
	if (nosbi)
		wrcltimecmp(VMASK(63));		/* no clock intrs on m */
}

/*
 * sys->machptr[cpu] must be set (thus a Mach allocated)
 * before calling this from low addresses.  on secondary cpus,
 * the MMU may be configured, but we're still in low addresses, and
 * could be in machine mode.
 */
static void
setmach(Sys *lowsys, uint cpu)
{
	Mach *m0;

	m = lowsys->machptr[cpu];	/* m must be a high address */
	if (m == nil)
		panic("setmach: nil sys->machptr[%d] before mallocinit", cpu);
	if (!bootmachmode)
		m = ensurelow(m);
	newmach(m, cpu);
	m0 = ensurelow(lowsys->machptr[0]);
	m->cpuhz  = m0->cpuhz;
	m->cpumhz = m0->cpumhz;
}

static void
alignchk(void)
{
	static int align = 0x123456;

	if (align != 0x123456)
		panic("data segment misaligned");
}

static void
zerosyssome(Sys *lowsys)
{
	char *aftreboot;

	/*
	 * zero Syspercpu, including machstk before use, for usage measurements.
	 */
	memset(&lowsys->Syspercpu, 0, sizeof(Syspercpu));

	/* zero Sys from syspage after Reboot to before kmesg */
	aftreboot = (char *)&lowsys->Reboot + sizeof(Reboot);
	memset(aftreboot, 0, (char *)lowsys->sysextend - aftreboot);
}

extern uint harts_per_cluster;

/*
 * On cpu0, initialise m & sys, allocate a permanent stack in sys->machstk,
 * return stack top in low addresses.
 */
static uintptr
setstkmach0(Sys *lowsys)
{
	int i;
	ulong ktzerophys, physmem;
	uintptr lowktzero;

	prf("--\n\n");
	lowsys->satp = 0;		/* don't use previous kernel's map */
	sys = KADDR((uintptr)lowsys);

	lowktzero = PADDR((void *)KTZERO);
	if (PPN(mainpc) != lowktzero)
		prf("_main pc %#p too far from PADDR(KTZERO) %#p, "
			"kernel loaded at wrong address!\n", mainpc, lowktzero);
	ktzerophys = ROUNDDN((ulong)lowktzero, GB);
	physmem = ROUNDDN(PHYSMEM, GB);
	if (physmem != ktzerophys) {
		prf("physmem %#lux != base of ktzero phys %#lux!\n",
			physmem, ktzerophys);
		prf("kernel loaded at address it was not linked for!\n");
	}
	if (0)
		prf("SB %#p at entry for _main %#p\n", getsb(), _main);

	alignchk();
	zerosyssome(lowsys);
	/* see loadids, which notes extensions */
	lowsys->nprivmodes = 2;	/* for plic: Machine & Super can take intrs */
	/* if hyper extension present, lowsys->nprivmodes++ */
	/* if user-interrupt extension present, lowsys->nprivmodes++ */

//	lowsys->machptr[0] = &sys->mach; /* probably unneeded; high address */
	newmach(&lowsys->mach, 0);	/* sets m to cpu0's (low) Mach */
	m->color = 0;			/* m->color = corecolor(0); */
	if (m->hartid - soc.hobbled >= harts_per_cluster)
		m->color = (m->hartid - soc.hobbled) / harts_per_cluster;

	if (initarchp) {
		initarchp = ensurelow(initarchp);
		dbprf("for %s cpus...", ensurelow(initarchp->name));
	}

	if (!bootmachmode && !nosbi) {
		/* dies on rvb-ice: no sbi, i guess */
		dbprf("cpu0: sbiget*...");
		lowsys->archid = sbigetmarchid();
		lowsys->vendorid = sbigetmvendorid();
	}

	/* try to detect a watchdog */
	if (TODO && Diagnose && Lowdebug) {
		dbprf("watchdog hanging over our heads? ");
		for (i = 0; i < 20; i++) {
			delay(500);
			dbprf(".");
		}
		dbprf("apparently not.\n");
	}

	return (uintptr)&lowsys->machstk[MACHSTKSZ];
}

static void
consputc(ulong *addr, uchar c)
{
	delay(1);
	*addr = c;
	coherence();
}

/*
 * the c910 doesn't like something about the way that I enable paging, but its
 * diagnostic is obscure (wandering off into the weeds) and it's generally
 * buggy.  other cpus and tinyemu think it's fine for Sv39 and Sv48.
 */
vlong
putsatp(uintptr satp)
{
	vlong ret;
	uintptr oldsatp;
	Mpl s;

	oldsatp = getsatp();
	if (oldsatp == satp)
		return 0;			/* no op */
	s = splhi();
	if (!m->machmode)
		if (oldsatp && satp == 0)
			assertrv("turning off mmu");
		else if (satp) {
			if (tlbinvall)
				tlbinvall();
			if (asids > 1)
				satp |= 1LL << Satpasidshft;
			assertrv("turning on mmu");
		}
	ret = _putsatp(satp);
	if (ret != 0)
		panic("cpu (or its configuration) is broken.  setting satp: "
			"exc cause %lld tval %#p epc %#p",
			ret, satptval, satpepc);
	if (Lowdebug && satp != 0) {
		/* try to verify new mapping */
		consputc((ulong *)PAUart0, '|');	/* works in s mode */
		if (!m->machmode)
			/* fails in s mode */
			consputc(ensurehigh(PAUart0), '#');
	}
	splx(s);
	return ret;
}

static void
stagger(int cpu)
{
	if (Lowdebug)
		delay(100 * cpu);
	USED(cpu);
}

/*
 * on a secondary cpu, secstall waits until stall is released, sets m,
 * waits for go-ahead, returns stack top.
 * another cpu must have allocated a Mach for us previously.
 */
static uintptr
secstall(Sys *lowsys, int cpu)
{
	stagger(cpu);			/* stagger startups for prints */
	dbprf("cpu%d stalling...", cpu);
	/*
	 * wait for cpu0 to allocate secondaries' data structures.
	 * the stall is cleared by schedcpus, settrampargs from reboot.
	 */
	while (lowsys->secstall)
		pause();

	coherence();
	stagger(cpu);			/* stagger startups for prints */
	/* sys->machptr[cpu] is now valid */
	if (cpu == 0)
		dbprf("cpu is 0 in secstall after stall!\n");
	setmach(lowsys, cpu);

	dbprf("cpu%d waiting for online...", cpu);
	USED(cpu);
	m->waiting = 1;			/* notify schedcpus that we are up */
	coherence();
	while (!m->online)
		pause();
	return m->stack + MACHSTKSZ;
}

static void
prfpc(int dummy)
{
	dbprf("pc %#p...", getcallerpc(&dummy));
	USED(dummy);
}

/*
 * start with the shared initial page table in lowsys->satp,
 * which should be lowsys->initpt.
 * enabling paging also enables speculative execution on xuantie.
 *
 * a netbsd porter says setting satp always causes a fault (S or M?);
 * it seems to make only the xuantie cpus lose their minds.
 *
 * this must not cause a page fault since up is still nil
 * (i.e., there is no user context).
 */
static void
pagingon(Sys *lowsys)
{
	uvlong satp;

	dbprf("set stvec for exten bugs %#p...", rectrapalign);
	putstvec(rectrapalign);	/* rectrapalign is low here when PC is low */
	if (initarchp && initarchp->supermppaging) {
		dbprf("cpu%d paging extension...", m->machno);
		(*(Pvfns)ensurelow(initarchp->supermppaging))(lowsys);
	}
	prfpc(0);
	dbprf("set stvec for kernel %#p...", ensurelow(strap));
	putstvec(ensurelow(strap));		/* va; needed? */
	if (m->machno == 0)
		mkinitpgtbl(lowsys);

	dbprf("switch to initial page tbl: writing satp %#p\n", lowsys->satp);
/**/	if (putsatp(lowsys->satp) < 0) {
		dbprf("satp %#p\n", getsatp());
		prfpc(0);
		panic("putsatp faulted on %#p", lowsys->satp);
	}
	dbprf("set stvec for kernel %#p...", ensurehigh(strap));
	putstvec(ensurehigh(strap));		/* va; needed? */
	prfpc(0);
	dbprf("\n");
	satp = getsatp();		/* asid field will be (# asids) - 1 */
	dbprf("read satp back as %#p\n", satp);
	if ((satp & ~Satpasidmask) != lowsys->satp)
		panic("couldn't enable paging with satp %#p", lowsys->satp);
}

static void
prstackuse(void)
{
	char *stkbase;
	uvlong *stkbot;

	if (Measurestkuse) {
		stkbase = initstks[m->machno];
		stkbot = (uvlong *)stkbase;
		prf("\ncpu%d: initstk %#p: %#p %#p %#p %#p use %lld bytes\n\n",
			m->machno, stkbase, stkbot[0], stkbot[1], stkbot[2],
			stkbot[3], stackuse(stkbase, stkbase + INITSTKSIZE));
	}
}

/* must be called with high addresses mapped */
static uintptr
prepstk(uintptr stktop)
{
	ulong wd;

	prstackuse();
	dbprf("high uart @ %#p: ", ensurehigh((void *)soc.uart));
	if (Lowdebug)
		/* dies if we started in S mode */
		*(ulong *)ensurehigh((void *)soc.uart) = '!';

	dbprf("new stack...");
	dbprf("\nlow stack %#p: writing...", stktop);
	*(ulong *)stktop = 0;
	coherence();

	stktop = (uintptr)ensurehigh((void *)stktop);
	dbprf("high stack %#p: reading...", stktop);
	wd = *(ulong *)stktop;
	coherence();

	dbprf("writing...");
	*(ulong *)stktop = 0;		/* 0 was: wd + 1 */
	USED(wd);
	coherence();

	dbprf("ok\n");
	return stktop;
}

/* tell other cpus to wait in secstall() */
static void
secstallset(Sys *lowsys)
{
	static int stallnext = RBFLAGSTALL;

	lowsys->secstall |= stallnext;
	stallnext = 0;				/* harmless race here */
	coherence();
}

void
setsts(void)
{
	/* S mode: little endian, prev mode = user, allow intrs in U mode */
	putsts((getsts() & ~(Ube|Spp|Uxl|Uie|Mpp|Sbe64)) |
		Spie | Defssts | Mppuser);
}

/*
 * We are called with all interrupts disabled.
 *
 * A temporary stack (in initstks) must be in use when we are called.
 * BSS has already been zeroed.  Don't use any high addresses until we are
 * definitely in supervisor mode with paging on.
 *
 * Main goals are: establish a permanent stack, m, sys, [sm]scratch, [sm]tvec,
 * [sm]status, page table, ensure cpu is in super mode, call main with machno.
 * Other cpus may be doing the same.
 *
 * sys is a shared global ptr, always a high address until shutdown, but our
 * use here will be of its (low) physical address, lowsys.  It's 1MB at the
 * top of the first memory bank.
 */
void
low(uint cpu)
{
	uintptr stktop;
	Sys *lowsys;

	setsts();
	if (cpu == 0)
		soc.clintlongs = 1;		/* pessimism, works on all */
	lowsys = (Sys *)(membanks[0].addr + membanks[0].size - MB);
	secstallset(lowsys);
	if (cpu >= MACHMAX) {
		prf("low: cpu%d out of range, >= %d; ignoring\n", cpu, MACHMAX);
		wfi();
	}

	/*
	 * On secondary cpus, wait for cpu0 to clear the stall in main.c.
	 * On all cpus, set m & sys, work out new stack top.
	 */
	stktop = cpu == 0? setstkmach0(lowsys): secstall(lowsys, cpu);

	/*
	 * on secondaries, m->online is now 1, so iovmapsoc must have run on
	 * cpu0 by now, thus its page table's kernel mappings are final.
	 * On all cpus, get out of machine mode if necessary, and start paging.
	 */
	dbprf("[cpu%d hart%d]\n", cpu, hartids[cpu]); /* show signs of life */
	if (bootmachmode)
		mach_to_super(lowsys);	/* enables any M extensions too */
	m->machmode = 0;

	dbprf("\n * supervisor mode.  sbi %s\n", nosbi? "not used": "assumed");
	if (0 && cpu == 0)
		prf("SB before pagingon %#p\n", getsb());
	pagingon(lowsys);
	csrswap(SCOUNTEREN, ~0ull);

	/*
	 * switch to new, permanent stack for this cpu at upper addresses.
	 * All extant automatic variables are on the old (current) stack, thus
	 * unpredictable after the switch.  The initstk can be reclaimed then.
	 */
	putstvec(rectrapalign);
/**/	setsp(prepstk(stktop - SBIALIGN));

	/*
	 * jump into high (kernel) addresses, thus vacating the lower range
	 * for user processes.  adjust pointers and registers as needed.
	 * gohigh calls jumphigh but may return to low saved return address
	 * on stack.
	 */
	dbprf("jump to kernel space...");
	gohigh();	/* sets up high straps, but returns to low address */
	jumphigh();	/* call asm directly: force regs high before main */
	prfpc(0);
	early = 0;

	dbprf("main...\n");
	main(m->machno);
	notreached();
}
