/*
 * prepare to start risc-v CPUs other than 0.
 *
 * unlike on x86, the harts all start at boot time.  we still have to
 * allocate memory for the secondary harts, so we make them wait until
 * we're ready for them to proceed.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"

#include "riscv.h"

/*
 * Allocate and minimally populate new page tables & Mach for cpu.
 *
 * The data needed per-processor is the sum of the stack, page
 * table pages, vsvm page and the Mach page.  The layout (Syscomm) is
 * similar to that described in dat.h for the bootstrap processor
 * (Sys), but with any unused space elided; see start.s, dat.h.
 *
 * NOTE: for now, share the page tables with the
 * bootstrap processor, until this code is worked out,
 * so only the Mach and stack portions are used below.
 */
static Mach *
newcpupages(int machno)
{
	Mach *mach;
	Syscomm *nsys;

	nsys = mallocalign(sizeof(Syscomm), PGSZ, 0, 0);
	if(nsys == nil) {
		print("newcpupages: no memory for cpu%d\n", machno);
		return nil;
	}

	/*
	 * Committed. If the cpu startup fails, can't safely
	 * release the resources, who knows what mischief
	 * the cpu is up to.  Perhaps should try to put it
	 * back into reset state?
	 */
	mach = &nsys->mach;
	mach->machno = machno;
	/* stack must immediately precede ptroot.  see dat.h. */
	mach->stack = (uintptr)nsys;
	mach->ptroot = m->ptroot;  /* share top-level page tables to start */
	sys->machptr[machno] = mach;	/* make Mach visible to new cpu */
	coherence();
	return mach;
}

void
cpusalloc(int ncpus)
{
	int machno;

	USED(ncpus);
	if (MACHMAX == 1)
		return;

	if (!nosbi)
		sbiclearipi();

	/* allocate memory for all cpus other than 0 */
	for(machno = 1; machno < MACHMAX && machno < ncpus+2; machno++)
		newcpupages(machno);
}
