/*
 * #P/cputype, shutdown and timing.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "../port/error.h"
#include "riscv.h"

#include <ctype.h>

enum {
	Qdir = 0,
	Qbase,

	Qmax = 16,

	Ghz	= 1000*1000*1000,

	Minclints = 20,			/* was 10 */
};

typedef long Rdwrfn(Chan*, void*, long, vlong);

static Rdwrfn *readfn[Qmax];
static Rdwrfn *writefn[Qmax];

static Dirtab archdir[Qmax] = {
	".",		{ Qdir, 0, QTDIR },	0,	0555,
};
Lock archwlock;	/* the lock is only for changing archdir */
int narchdir = Qbase;

uvlong cpuhz;

extern Dev archdevtab;

/*
 * Add a file to the #P listing.  Once added, you can't delete it.
 * You can't add a file with the same name as one already there,
 * and you get a pointer to the Dirtab entry so you can do things
 * like change the Qid version.  Changing the Qid path is disallowed.
 */
Dirtab*
addarchfile(char *name, int perm, Rdwrfn *rdfn, Rdwrfn *wrfn)
{
	int i;
	Dirtab d;
	Dirtab *dp;

	memset(&d, 0, sizeof d);
	strcpy(d.name, name);
	d.perm = perm;

	lock(&archwlock);
	if(narchdir >= Qmax){
		unlock(&archwlock);
		return nil;
	}

	for(i=0; i<narchdir; i++)
		if(strcmp(archdir[i].name, name) == 0){
			unlock(&archwlock);
			return nil;
		}

	d.qid.path = narchdir;
	archdir[narchdir] = d;
	readfn[narchdir] = rdfn;
	writefn[narchdir] = wrfn;
	dp = &archdir[narchdir++];
	unlock(&archwlock);
	return dp;
}

static Chan*
archattach(char* spec)
{
	return devattach(archdevtab.dc, spec);
}

Walkqid*
archwalk(Chan* c, Chan *nc, char** name, int nname)
{
	return devwalk(c, nc, name, nname, archdir, narchdir, devgen);
}

static long
archstat(Chan* c, uchar* dp, long n)
{
	return devstat(c, dp, n, archdir, narchdir, devgen);
}

static Chan*
archopen(Chan* c, int omode)
{
	return devopen(c, omode, archdir, narchdir, devgen);
}

static void
archclose(Chan*)
{
}

enum
{
	Linelen= 31,
};

static long
archread(Chan *c, void *a, long n, vlong offset)
{
	Rdwrfn *fn;

	switch((ulong)c->qid.path){

	case Qdir:
		return devdirread(c, a, n, archdir, narchdir, devgen);
	default:
		if(c->qid.path < narchdir && (fn = readfn[c->qid.path]))
			return fn(c, a, n, offset);
		error(Eperm);
		notreached();
	}
}

static long
archwrite(Chan *c, void *a, long n, vlong offset)
{
	Rdwrfn *fn;

	switch((ulong)c->qid.path){
	default:
		if(c->qid.path < narchdir && (fn = writefn[c->qid.path]))
			return fn(c, a, n, offset);
		error(Eperm);
		notreached();
	}
}

Dev archdevtab = {
	'P',
	"arch",

	devreset,
	devinit,
	devshutdown,
	archattach,
	archwalk,
	archstat,
	archopen,
	devcreate,
	archclose,
	archread,
	devbread,
	archwrite,
	devbwrite,
	devremove,
	devwstat,
};

void
nop(void)
{
}

static long
cputyperead(Chan*, void *a, long n, vlong off)
{
	char str[32];

	snprint(str, sizeof(str), "%s %ud\n", cputype, m->cpumhz);
	return readstr(off, a, n, str);
}

static ulong exts[] = {
	0, 1, 2, 3, 4, 5, 6, 7, 8, 0x10,
	0x735049, 0x52464e43, HSM, 0x53525354,
};

static void
sbiext(int ext)
{
	char c;

	if (sbiprobeext(ext) != 0) {
		print(" %#ux", ext);
		if (ext >= 256) {
			print(" (");
			for (; ext > 0; ext <<= 8) {
				c = ((uchar *)&ext)[3];
				if (c && isprint(c))
					print("%c", c);
			}
			print(")");
		}
	}
}

static void
sbiinit(void)
{
	int i;

	if (machine_mode)
		/* reads back as 0x222 (Ssie|Stie|Seie) on tinyemu */
		print("mideleg %#ux\n", mideleg); /* varies with implement'n */
	else if (!nosbi) {
		havesbihsm = sbiprobeext(HSM) != 0;
		print("sbi extensions present:");
		for (i = 0; i < nelem(exts); i++)
			sbiext(exts[i]);
		print("\n");
	}
}

void
archinit(void)
{
	addarchfile("cputype", 0444, cputyperead, nil);
	sbiinit();
}

void
archreset(void)
{
	Htif *htif = (Htif *)htifpwrregs;	/* tinyemu only */

	cacheflush();
	spllo();
	if (htif) {
		iprint("htif resetting.\n");
		htif->tohost[0] = 1;
		htif->tohost[1] = 0;		/* cmd 1: power off */
	}
	delay(1);
	if (!nosbi && m->machno == 0) {
		iprint("sbi shutting down.\n");
		sbishutdown();			/* shuts down all cpus */
	}
	iprint("last resort: into wfi.\n");
	for (;;)
		halt();
}

static void
haltsys(void)
{
	int i;

	iprint("mpshutdown: %d active cpus\n", sys->nonline);
	delay(5);
	devtabshutdown();
	splhi();

	/* paranoia: wait for other cpus */
	for (i = 100; i > 0 && sys->nonline > 1; i--) {
		iprint(" %d", sys->nonline);
		delay(200);
	}

	/* we can now use the uniprocessor reset */
	archreset();
	notreached();
}

/*
 * secondary cpus should all idle (here or in shutdown, called from reboot).
 * for a true shutdown, reset and boot, cpu 0 should then reset itself or the
 * system as a whole.  if plan 9 is rebooting (loading a new kernel and
 * starting it), cpu 0 needs to keep running to do that work, so we return
 * in that case only.
 *
 * we assume shutdown has previously been called on this cpu.
 */
void
mpshutdown(void)
{
	void (*tramp)(Reboot *);

	if(m->machno == 0) {
		static QLock mpshutdownlock;

		qlock(&mpshutdownlock);
		/* we are the first proc on cpu0 to begin shutdown */
		if(!active.rebooting) {
			haltsys();
			notreached();
		}
	} else {
		/*
		 * we're secondary, we should go into sys->reboottramp to enter
		 * wfi.  if idling, sys->Reboot won't be set and that's okay.
		 */
		cacheflush();
		mmulowidmap();			/* disables user mapping too */
		m->virtdevaddrs = 0;
		/* await populated sys->Reboot */
		while (sys->secstall == RBFLAGSET)
			coherence();
		tramp = (void (*)(Reboot *))PADDR(sys->reboottramp);
		(*tramp)(&sys->Reboot);
		notreached();
	}
}

/*
 * there are two main timers: the clint clock (mtime)
 * and the cpu clock (rdtsc).  The clint clock is typically
 * something like 1MHz and the cpu clock 600MHz.
 */

/*
 *  return value and speed of timer
 */
uvlong
fastticks(uvlong* hz)
{
	if(hz != nil)
		*hz = cpuhz;
	return rdtsc();
}

ulong
µs(void)
{
	return fastticks2us(fastticks(nil));
}

void
setclinttmr(Clint *clnt, uvlong clints)
{
	/* seems not to work under OpenSBI, though restriction is undocumented */
	clnt->mtimecmp[m->hartid] = clints;
	coherence();

	if (!nosbi)
		setsbitimer(clints);
}

/*
 * return cpu cycles / clint tick ratio.
 * u54-mc clint (rtc) clock must run at less than 1/2 the cycle rate.
 */
uvlong
calibrate(void)
{
	uvlong tsc, mtime, tsc2, mtime2, tscperclint;
	Clint *clint;

	if (sys->tscperclint)
		return sys->tscperclint;
	clint = clintaddr();
	clint->mtime = 0;		/* reset while only 1 cpu running */
	coherence();
	mtime = clint->mtime;		/* in clint units */
	tsc = rdtsc();			/* in cycles */
	delay(1);
	mtime2 = clint->mtime;
	tsc2 = rdtsc();

	tscperclint = 1;
	if (mtime2 > mtime) {
		/*
		 * no need to round: a lower tscperclint will cause larger
		 * tick delays in timerset, which should be harmless.
		 */
		tscperclint = (tsc2 - tsc) / (mtime2 - mtime);
		if ((vlong)tscperclint < 1)
			tscperclint = 1;
	}
	print("%,llud clint ticks is %,llud rdtsc cycs, or %lld cycs/clint\n",
		mtime2 - mtime, tsc2 - tsc, tscperclint);
	sys->clintsperhz = timebase / HZ;	/* clint ticks per HZ */
	sys->minclints = 20 * timebase / 1000000; /* min. interval until intr */
	return tscperclint;
}

void
clockoff(Clint *clint)
{
	setclinttmr(clint, VMASK(63));
	clrstip();
}

/*
 *  set next timer interrupt at time next, in cycles (fastticks).
 *  as a special case, next==0 requests a minimal interval.
 */
void
timerset(uvlong next)
{
	Mpl pl;
	vlong cycs, clints, tscperclint;
	Clint *clint;

	pl = splhi();
	clints = sys->minclints;
	if (next != 0) {
		cycs = next - fastticks(nil);
		tscperclint = sys->tscperclint;
		if (tscperclint != 0 && cycs > tscperclint)
			clints = cycs / tscperclint;

		/* enforce sane bounds */
		if (clints > sys->clintsperhz)
			clints = sys->clintsperhz;
		else if (clints < sys->minclints)
			clints  = sys->minclints;
	}
	if (clints < Minclints)		/* @ 1MHz clint, 1 clint is 1 µs */
		clints = Minclints;	/* don't interrupt immediately */

	clint = clintaddr();
	clockoff(clint);		/* don't trigger early */
	setclinttmr(clint, clint->mtime + clints);
	clockenable();
	splx(pl);
}
