/*
 * assembly-language machine assist for riscv rv32.
 *
 * never use TEXT ... $4; use $8 to keep stack vlong-aligned for ic.
 */
#include "mem.h"
#include "riscvl.h"

/*
 * Read/write various system registers.
 */

/* invoked as `t = rdtsc()'.  &t is passed in R(ARG). */
TEXT rdtsc(SB), $-4				/* Time Stamp Counter */
again:
	MOV	CSR(CYCHI), R9
	MOV	CSR(CYCLO), R11
	MOV	CSR(CYCHI), R10
	BNE	R9, R10, again			/* wrapped? try again */
	MOV	R11, 0(R(ARG))
	MOV	R9, XLEN(R(ARG))
	RET

TEXT getsp(SB), $-4
	MOV	R2, R(ARG)
	RET
TEXT getsts(SB), 1, $-4
	MOV	CSR(SSTATUS), R(ARG)
	RET
TEXT putsts(SB), 1, $-4
	MOV	R(ARG), CSR(SSTATUS)
	RET
TEXT getsip(SB), $-4
	MOV	CSR(SIP), R(ARG)
	RET
TEXT setsb(SB), $-4
	MOV	$setSB(SB), R3
	RET

TEXT clockenable(SB), $-4	
	MOV	$Stie, R9			/* super timer intr enable */
	CSRRS	CSR(SIE), R9, R0
	RET

TEXT clrstip(SB), $-4
	MOV	$Stie, R9
	CSRRC	CSR(SIP), R9, R0
	CSRRC	CSR(SIE), R9, R0
	RET

TEXT clrsipbit(SB), $-4
	CSRRC	CSR(SIP), R(ARG), R0
	RET

TEXT getsatp(SB), $-4
	MOV	CSR(SATP), R(ARG)
	RET
TEXT putsatp(SB), $-4
	FENCE
	SFENCE_VMA(0, 0)			/* flush TLB */
	MOV	R(ARG), CSR(SATP)
	SFENCE_VMA(0, 0)			/* flush TLB */
	RET

/*
 * cache or tlb invalidation
 */
TEXT invlpg(SB), 1, $-4
	FENCE
	FENCE_I
	/* hifive unmatched can't do a non-global SFENCE */
//	SFENCE_VMA(0, ARG)	/* only fences pt leaf accesses for va in ARG */
	SFENCE_VMA(0, 0)
	RET

/* approximate intel's wbinvd instruction: i & d fences, flush mmu state */
TEXT wbinvd(SB), 1, $-4
	FENCE
	FENCE_I
	SFENCE_VMA(0, 0)
	RET

TEXT cacheflush(SB), 1, $-4
	FENCE
	FENCE_I
	RET

/*
 * Serialisation.
 */
TEXT coherence(SB), 1, $-4
	FENCE
	RET

TEXT splhi(SB), 1, $-4
_splhi:
	FENCE
	CSRRC	CSR(SSTATUS), $Sie, R(ARG) /* clear ie bits */
	AND	$Sie, R(ARG)
	BEQ	R(ARG), _spldone	/* Sie=0, intrs were disabled? */
	MOV	LINK, 4(R(MACH)) 	/* save PC in m->splpc */
_spldone:
	RET

TEXT spllo(SB), 1, $-4			/* marker for devkprof */
_spllo:
	CSRRS	CSR(SSTATUS), $Sie, R(ARG) /* set ie bits */
	AND	$Sie, R(ARG)
	BNE	R(ARG), _spldone	/* Sie=1, intrs were enabled? */
	MOV	R0, 4(R(MACH))		/* clear m->splpc */
	FENCE
	RET

/* assumed between spllo and spldone by devkprof */
TEXT splx(SB), 1, $-4
	MOV	CSR(SSTATUS), R9
	AND	$Sie, R9
	AND	$Sie, R(ARG), R10
	BEQ	R9, R10, _spldone	/* already in desired state? */
	BEQ	R10, _splhi		/* want intrs disabled? */
	JMP	_spllo			/* want intrs enabled */

TEXT spldone(SB), 1, $-4		/* marker for devkprof */
	RET

TEXT islo(SB), 1, $-4
	MOV	CSR(SSTATUS), R(ARG)
	AND	$Sie, R(ARG)
	RET

/*
 * atomic operations
 */

TEXT aincnonneg(SB), 1, $8			/* int aincnonneg(int*); */
	JAL	LINK, ainc(SB)
	BGT	R(ARG), _aincdone
	/* <= 0 after incr, so overflowed, so blow up */
	MOV	(R0), R0			/* over under sideways down */
_aincdone:
	RET

TEXT adecnonneg(SB), 1, $8			/* int adecnonneg(int*); */
	JAL	LINK, adec(SB)
	BGE	R(ARG), _aincdone
	/* <0, so underflow.  use own copy of XORQ/MOVQ to disambiguate PC */
	MOV	(R0), R0			/* over under sideways down */
	RET

/* see libc/riscv/atom.s for ainc & adec */
TEXT aadd(SB), 1, $-4			/* int aadd(int*, int); */
	MOV	R(ARG), R12		/* address of counter */
	MOVW	addend+XLEN(FP), R10
	FENCE
loop:
	MOVW	R10, R13
	LRW(12, ARG)			/* (R12) -> R(ARG) */
	ADDW	R(ARG), R13
	MOVW	R13, R(ARG)		/* return new value */
	SCW(13, 12, 14)			/* R13 -> (R12) maybe, R14=0 if ok */
	BNE	R14, loop
	RET

/*
 * oddball instruction.
 */
TEXT ecall(SB), 0, $-4
	ECALL
	RET

/*
 * idle until an interrupt.
 * really only useful on old, uniprocessor VMs.  returns spllo().
 */
TEXT halt(SB), 1, $-4
	CSRRC	CSR(SSTATUS), $Sie, R10		/* splhi: clear ie bit */
	MOV	nrdy(SB), R9
	BNE	R9, _procready
#ifndef TINYEMU
	/* WFI seems to hang in tinyemu after several minutes of up time. */
	WFI		/* will be awakened by an intr, even if disabled */
#endif
_procready:
	CSRRS	CSR(SSTATUS), $Sie, R0		/* spllo: set ie bit */
	RET

TEXT reset(SB), 1, $-4
bragain:
	EBREAK
	JMP	bragain

/*
 * Label consists of a stack pointer and a programme counter.
 */
TEXT gotolabel(SB), 1, $-4			/* gotolabel(Label *) */
	MOV	0(R(ARG)), R2			/* restore sp */
	MOV	XLEN(R(ARG)), LINK		/* restore return pc */
	MOV	$1, R(ARG)
	RET

TEXT setlabel(SB), 1, $-4			/* setlabel(Label *) */
	MOV	R2, 0(R(ARG))			/* store sp */
	MOV	LINK, XLEN(R(ARG))		/* store return pc */
	MOV	R0, R(ARG)
	RET

#ifdef TODO				/* untested optimised RV32 mul64fract */
/*
 * mul64fract(uvlong*r, uvlong a, uvlong b)
 * Multiply two 64 numbers and return the middle 64 bits of the 128 bit result.
 *
 *	ulong ah, al, bh, bl;
 *	bl = b; bh = b >> 32; al = a; ah = a >> 32;
 *	*r = (((uvlong)al*bl)>>32) + (uvlong)al*bh + (uvlong)ah*bl +
 *		(((uvlong)ah*bh)<<32);
 *
 *	hold *r in progress in R16, R15
 */
TEXT mul64fract(SB), 1, $-4
	/* r, the address of the result, is in RARG */
	MOV	a+4(FP), R9			/* lo */
	MOV	a+8(FP), R10			/* hi */
	MOV	b+12(FP), R11			/* lo */
	MOV	b+16(FP), R12			/* hi */

	MULHU	R9, R11, R13			/* high (al*bl) */
//	MUL	R9, R12, R14			/* low (al*bl): discard */

	MULHU	R9, R12, R15			/* high (al*bh) */
	MUL	R9, R12, R17			/* low (al*bh) */

	ADD	R13, R17, R16			/* sum low */
	BGEU	R16, R13, nooverflow1		/* risc-v uns overflow test */
	ADD	$1, R15				/* add carry to high */
nooverflow1:

	MULHU	R9, R12, R13			/* high (ah*bl) */
	MUL	R9, R12, R14			/* low (ah*bl) */

	ADD	R14, R16, R17			/* sum low */
	BGEU	R17, R14, nooverflow2		/* risc-v uns overflow test */
	ADD	$1, R15				/* add carry to high */
nooverflow2:
	ADD	R13, R15			/* sum high */
	MOV	R17, R16			/* put low back in R16 */

//	MULHU	R9, R11, R13			/* high (ah*bh): discard */
	MUL	R9, R12, R14			/* low (ah*bh) */

	ADD	R14, R15			/* low += high */

	MOV	R16, (R(ARG))			/* store low */
	MOV	R15, 4(R(ARG))			/* store high */
	RET
#endif
