/*
 * initialise a risc-v RV32G system (if any ever exist), start all cpus, and
 * start scheduling processes on them, notably /boot/boot as process 1.
 * also contains graceful shutdown and reboot code.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"

#include "io.h"
#include "ureg.h"
#include "riscv.h"

#include "init.h"
#include "reboot.h"
#include <a.out.h>
#include <ctype.h>

enum {
	Ps2 = 0,	/* flag: allow PS/2 keyboard */
};

extern uintptr membanks[];
extern char defnvram[];

#ifdef MACHINE_MODE
int machine_mode = 1;
#else
int machine_mode = 0;
#endif

extern void ecall(...);

/* mmu.c */
extern void pageidmap(void);
extern uintptr topram(void);

static uintptr usp = 0;		/* user stack of init proc */

Sys* sys = nil;			/* initializer keeps sys out of bss */
usize sizeofSys = sizeof(Sys);
usize sizeofSyscomm = sizeof(Syscomm);

/*
 * Option arguments from the command line.
 * oargv[0] is the boot file (/boot/boot).
 * mboptinit() was called from multiboot() on amd64 to set it all up.
 */
static int oargc;
static char* oargv[20];
static char oargb[128];
static int oargblen;

char cant[] = "can't happen: notreached reached\n";
char dbgflg[256] = { 0, };
char cputype[] = "riscv";

static int vflag = 0;

#ifdef TINYEMU
void
pcireset(void)
{
}
#endif

void
pause(void)
{
}

/* dreg from parsing multiboot options on pcs */
void
mboptinit(char* s)
{
	oargblen = strecpy(oargb, oargb+sizeof(oargb), s) - oargb;
	oargc = tokenize(oargb, oargv, nelem(oargv)-1);
	oargv[oargc] = nil;
}

static void
fakecpuhz(void)
{
	/*
	 * Need something for initial delays
	 * until a timebase is worked out.
	 */
	m->cpuhz = 2000000000ll;
	m->cpumhz = 2000;
	m->perf.period = 1;
}

static vlong
sethz(void)
{
	vlong hz;

	hz = archhz();
	if(hz != 0){
		m->cpuhz = hz;
		m->cpumhz = hz/1000000;
	}
	return hz;
}

static void
setepochtsc(void)
{
	uvlong now;

	now = rdtsc();
	m->rdtsc = now;
	if (m->machno == 0)
		sys->epoch = now;
}

/*
 * start a cpu other than 0.  entered splhi.
 * must copy root pt to a newly-allocated one; see cpustart.c.
 */
void
squidboy(uint machno)
{
	up = nil;
	if (machno >= MACHMAX) {
		print("cpu #%d >= MACHMAX (%d)\n", machno, MACHMAX);
		return;
	}
	m = sys->machptr[machno];	/* already set by newcpupages() */
	m->machno = machno;
	fakecpuhz();

	DBG("Hello Squidboy %d\n", m->machno);

	/*
	 * Beware the Curse of The Non-Interruptable Were-Temporary.
	 */
	if(sethz() == 0) {
		print("squidboy: cpu%d: 0Hz\n", m->machno);
		notreached();
	}

	mmuinit();
	fpuinit();

	/*
	 * Handshake with cpu to let it know the startup succeeded.
	 */
	m->splpc = 0;
	coherence();

	/*
	 * Handshake with main to proceed with initialisation.
	 */
	while(sys->epoch == 0)
		pause();
	setepochtsc();

	DBG("mach %d is go %#p %#p\n", m->machno, m, m->ptroot->va);
	/*
	 * we used to switch on m->mode here, but nothing in the current
	 * system sets it.
	 */
	timersinit();

	/*
	 * Cannot allow interrupts while waiting for online.
	 * A clock interrupt here might call the scheduler,
	 * and that would be a mistake.
	 */
	while(!m->online)
		pause();
	clockenable();

	DBG("mach%d: online color %d\n", m->machno, m->color);
	schedinit();			/* no return */
	panic("cpu%d: schedinit returned", m->machno);
}

/*
 * ilock via i8250enable via i8250console
 * needs m->machno, sys->machptr[] set, and
 * also 'up' set to nil.
 */
static void
machsysinit(void)
{
	memset(m, 0, sizeof(Mach));
	m->machno = 0;
	m->online = 1;
	sys->machptr[m->machno] = &sys->mach;
	m->stack = PTR2UINT(sys->machstk);
	sys->nmach = 1;
	sys->nonline = 0;
	cpuactive(0);
	sys->copymode = 0;			/* COW */
	fakecpuhz();
}

/*
 * Release the hounds.
 * Don't delay here; a 2 s. delay prevents interrupts on other cpus.
 */
static void
schedcpus(void)
{
	int i;
	Mach *mp;

	sys->secstall = 0;		/* let them start and poll m->online */
	for(i = 1; i < MACHMAX; i++){
		mp = sys->machptr[i];
		if(mp == nil)
			continue;

		lock(&active);
		cpuactive(i);
		unlock(&active);

		mp->color = 0;		// corecolor(i);
		if(mp->color < 0)
			mp->color = 0;
		mp->online = 1;
		coherence();
	}
	if (sys->nonline > 1)
		print("%d cpus running\n", sys->nonline);
}

uintptr kernmem;

/* set size of memory for page tables & qmalloc */
void
setkernmem(void)
{
	uintptr pmsize = sys->pmend - sys->pmbase;

	if (pmsize < 128*MB) {		/* sanity */
		print("mem size %,llud; assuming 128 MB of ram\n",
			(uvlong)pmsize);
		pmsize = 128*MB;
		sys->pmend = sys->pmbase + pmsize;
	}
	kernmem = physsize / 5;		/* arbitrary */
	if (kernmem >= pmsize/4)
		kernmem = ROUNDUP(pmsize/4, PGSZ);
}

static void
multiprocinit(void)
{
	cpusalloc(1);
	setepochtsc();
	/* when a cpu starts, it copies the page table root to its own copy */
	schedcpus();
}

/* first things after zeroing BSS */
static void
physmemconsopts(u32int mbmagic, u32int mbi)
{
	sys->pmbase = physmem;
	machsysinit();
	asminit();
	/* make asmlist & palloc.mem from multiboot memory map */
	/* have no multiboot loading on risc-v */
	if (0 && multiboot(mbmagic, mbi, 0) < 0)
		panic("no multiboot info");
	USED(mbmagic, mbi);
	/* a single bank of memory on icicle */
	asmmapinit(sys->pmbase, physsize - topram(), AsmMEMORY);
	/* we should know physical memory size (sys->pmend) here */

	mboptinit("/boot/ipconfig ether");	/* dreg from pc multiboot */

	i8250console("0 b115200");
	kmesginit();
//	cpuactive(0);		/* now done by machsysinit */
	active.exiting = 0;
}

/* also calls printinit for console input */
static void
configcpumem(int)
{
	/*
	 * Mmuinit before meminit because it makes mappings and flushes
	 * the TLB.  It will get page table pages from Sys as needed.
	 */
	setkernmem();
	mmuinit();	/* uses kernmem; sets sys->vm*; vmaps soc devices */

	if (Ps2)
		kbdinit();

	meminit();			/* map KSEG[02] to ram */
	archinit();			/* populate #P with cputype */
	if (!mallocok)
		mallocinit();
	trapinit();

	/*
	 * Printinit should cause the first malloc call to happen (printinit->
	 * qopen->malloc).  If the system dies here, it's probably due to malloc
	 * not being initialised correctly, or the data segment is misaligned
	 * (it's amazing how far you can get with things like that completely
	 * broken).
	 */
	printinit();	/* actually, establish cooked console input queue */

	pcireset();			/* turn off bus masters & intrs */
}

	uvlong	calibrate(void);
extern	void	zerotrapcnts(void);

/*
 * copy reboot trampoline function to its expected location.
 * a reboot could be requested at any time, and all cores will
 * jump to the trampoline.  some of them may still be there,
 * so only overwrite if necessary.
 */
static void
copyreboottramp(void)
{
	if (memcmp(sys->reboottramp, rebootcode, 250) != 0)
		memmove(sys->reboottramp, rebootcode, sizeof(rebootcode));
	cacheflush();
}

/*
 * entered in supervisor mode with paging off.
 * don't touch BSS data nor Sys until after memset zeroes them.
 */
void
main(int cpu)
{
	static ulong align = 1234;

	splhi();
	if (cpu != 0) {
		squidboy(cpu);
		notreached();
	}
	memset(edata, 0, end - edata);	/* zero BSS */
	setsb();
	/* zero Sys excluding stack & kmesg */
	memset((uchar *)sys + MACHSTKSZ, 0, LOW0ZEROEND - MACHSTKSZ);

	fmtinit();			/* install P, L, R, Q verbs */
	physmemconsopts(0, 0);
	/* we know physical memory size (sys->pmend) here */

	print("\n\nPlan 9 (risc-v RV%dG)\n", BI2BY*(int)sizeof(uintptr));
	if (align != 1234)
		panic("mis-aligned data segment");
	sanity();
	assert((getsp() % sizeof(vlong)) == 0);

	sethz();			/* prints cpu id */
	pageidmap();			/* enables mmu with identity map */

	sys->tscperclint = calibrate();
	configcpumem(0); /* configs mmu, vmaps soc devs, resets pci devs */

	timersinit();
	addclock0link(zerotrapcnts, 1000);
	kbdenable();		/* allocate kbd input queue */
	fpuinit();
	psinit();
	initimage();

	clocksanity();		/* enables the clock */
	links();
	devtabreset();		/* discover all devices */
	m->machmode = machine_mode;
	copyreboottramp();
	/*
	 * start all cpus to allow better intr assignments.  assumes no
	 * kprocs will be started (e.g., due to devtabreset) until schedinit
	 * runs init0 (see userinit).
	 *
	 * kernel memory mappings (including in VMAP) must be fixed when
	 * multiprocinit is called, since they will be copied for each cpu
	 * just before it starts.  drivers typically call vmap in their
	 * reset (`pnp') functions, called from devtabreset.
	 */
	multiprocinit();

	pageinit();		/* make non-kernel memory available for users */
	userinit();	/* hand-craft init process; needs demand paging */
	schedinit();	/* start user init process; no return */
	panic("cpu0 schedinit returned");
}

/*
 * process 1 runs this (see userinit), so it's now okay for devtabinit to
 * spawn kprocs.
 */
void
init0(void)
{
	char buf[2*KNAMELEN];

	up->nerrlab = 0;

	spllo();

	/*
	 * These are o.k. because rootinit is null.
	 * Then early kproc's will have a root and dot.
	 */
	up->slash = namec("#/", Atodir, 0, 0);
	pathclose(up->slash->path);
	up->slash->path = newpath("/");
	up->dot = cclone(up->slash);

	devtabinit();

	if(!waserror()){
		snprint(buf, sizeof(buf), "%s %s", cputype, conffile);
		ksetenv("terminal", buf, 0);
		ksetenv("cputype", cputype, 0);
		if(cpuserver)
			ksetenv("service", "cpu", 0);
		else
			ksetenv("service", "terminal", 0);
		ksetenv("nvram", defnvram, 0);
		ksetenv("nobootprompt", "tcp", 0);
		ksetenv("nobootprompt", "tcp", 1);	// TODO: remove
//		confsetenv();			/* no conf to convert */
		poperror();
	}
	kproc("alarm", alarmkproc, 0);
	/*
	 * start user phase executing initcode[] from init.h, compiled
	 * from init9.c (main) and ../port/initcode.c (startboot),
	 * which in turn execs /boot/boot.
	 *
	 * sp is a result of bootargs.
	 */
	assert((usp % sizeof(vlong)) == 0);
	touser(usp);
}

uintptr
bootargs(uintptr base)
{
	int i;
	uintptr ssize;
	char **av, *p;

	/*
	 * Push the boot args onto the stack.
	 * Make sure the validaddr check in syscall won't fail
	 * because there are fewer than the maximum number of
	 * args by subtracting sizeof(up->arg).
	 */
	i = oargblen+1;
	p = (void *)STACKALIGN(base + PGSZ - sizeof(up->arg) - i);
	memmove(p, oargb, i);

	/*
	 * Now push argc and the argv pointers.
	 * This isn't strictly correct as the code jumped to by
	 * touser in init9.[cs] calls startboot (port/initcode.c) which
	 * expects arguments
	 * 	startboot(char* argv0, char* argv[])
	 * not the usual (int argc, char* argv[]), but argv0 is
	 * unused so it doesn't matter (at the moment...).
	 *
	 * should add another +2 to oargc?  +1 so nil isn't stepped on,
	 * another for vlong padding.
	 */
	av = (char**)(p - (oargc+2)*sizeof(char*));
	ssize = base + PGSZ - PTR2UINT(av);
	*av++ = (char*)oargc;
	for(i = 0; i < oargc; i++)
		*av++ = (oargv[i] - oargb) + (p - base) + (USTKTOP - PGSZ);
	*av = nil;
	return USTKTOP - ssize;
}

void
userinit(void)
{
	Proc *p;
	Segment *s;
	KMap *k;
	Page *pg;

	p = newproc();
	p->pgrp = newpgrp();
	p->egrp = smalloc(sizeof(Egrp));
	p->egrp->ref = 1;
	p->fgrp = dupfgrp(nil);
	p->rgrp = newrgrp();
	p->procmode = 0640;

	kstrdup(&eve, "");
	kstrdup(&p->text, "*init*");
	kstrdup(&p->user, eve);

	/*
	 * Kernel Stack
	 *
	 * N.B. make sure there's enough space for syscall to check
	 *	for valid args and
	 *	space for gotolabel's return PC
	 */
	p->sched.pc = PTR2UINT(init0);	/* proc 1 starts here in kernel phase */
	p->sched.sp = PTR2UINT(p->kstack+KSTACK-sizeof(up->arg));
	p->sched.sp = STACKALIGN(p->sched.sp);

	/*
	 * User Stack
	 *
	 * Technically, newpage can't be called here because it
	 * should only be called when in a user context as it may
	 * try to sleep if there are no pages available, but that
	 * shouldn't be the case here.
	 */
	s = newseg(SG_STACK, USTKTOP-USTKSIZE, USTKTOP);
	p->seg[SSEG] = s;
	pg = newpage(Zeropage, s, USTKTOP-(1<<s->lg2pgsize), 0);
	segpage(s, pg);
	k = kmap(pg);
	usp = bootargs(VA(k));
	/* usp will be init0's stack pointer via touser */
	kunmap(k);

	/*
	 * Text
	 */
	s = newseg(SG_TEXT, UTZERO, UTZERO+PGSZ);
	s->flushme++;
	p->seg[TSEG] = s;
	pg = newpage(Zeropage, s, UTZERO, 0);
	memset(pg->cachectl, PG_TXTFLUSH, sizeof(pg->cachectl));
	segpage(s, pg);
	k = kmap(s->map[0]->pages[0]);
	memmove((void *)VA(k), initcode, sizeof initcode);
	kunmap(k);

	wbinvd();
	/* not using address-space ids currently */
	putsatp(Pagingon | (m->ptroot->pa / PGSZ));

	ready(p);
}

static void
drainuart(void)
{
	int i;

	if (!islo()) {
		iprint("drainuart: called splhi\n");
		delay(100);
		return;
	}
	for (i = 300; i > 0 && consactive(); i--)
		delay(10);
}

/* shutdown this cpu */
static void
shutdown(int ispanic)
{
	int ms, once;

	/* simplify life by shutting off any watchdog */
	if (watchdogon && watchdog) {
		watchdog->disable();
		watchdogon = 0;
	}

	lock(&active);
	if(ispanic)
		active.ispanic = ispanic;
	else if(m->machno == 0 && !iscpuactive(0))
		active.ispanic = 0;		/* reboot */
	once = iscpuactive(m->machno);
	/*
	 * setting exiting will make hzclock() on each processor call exit(0),
	 * which calls shutdown(0) and mpshutdown(), which idles non-bootstrap
	 * cpus and returns on bootstrap processors (to permit a reboot).
	 * clearing our bit in active.machsmap avoids calling exit(0) from
	 * hzclock() on this processor.
	 */
	cpuinactive(m->machno);
	active.exiting = 1;
	unlock(&active);

	if(once)
		iprint("cpu%d: %s...", m->machno, m->machno? "idling": "exiting");
	spllo();
	for(ms = 5*1000; ms > 0; ms -= TK2MS(2)){
		delay(TK2MS(2));
		if(sys->nonline <= 1 && consactive() == 0)
			break;
	}

	cacheflush();
	if(active.ispanic && m->machno == 0){
		devtabshutdown();
		if(cpuserver)
			delay(30000);
		else
			for(;;)
				idlehands();
	} else
		delay(1000);
}

void
exit(int ispanic)
{
	shutdown(ispanic);
	mpshutdown();		/* returns only if we are rebooting */
	archreset();
}

static int
okkernel(int magic)
{
	return magic == AOUT_MAGIC;
}

int (*isokkernel)(int) = okkernel;

/*
 * if we have to reschedule, up must be set (i.e., we must be in a
 * process context).
 */
void
runoncpu(int cpu)
{
	if (m->machno == cpu)
		return;			/* done! */
	if (up == nil)
		panic("runoncpu: nil up");
	if (up->nlocks)
		print("runoncpu: holding locks, so sched won't work\n");
	procwired(up, cpu);
	sched();
	if (m->machno != cpu)
		iprint("cpu%d: can't switch proc to cpu%d\n", m->machno, cpu);
}

static void
shutothercpus(void)
{
	/*
	 * the boot processor is cpu0.  execute this process on it
	 * so that the new kernel has the same cpu0.  this only matters
	 * because the hardware has a notion of which processor was the
	 * boot processor and we look at it at start up.
	 */
	if (m->machno != 0 && up)
		runoncpu(0);

	/*
	 * the other cpus could be holding locks that will never get
	 * released (e.g., in the print path) if we put them into
	 * reset now, so ask them to shutdown gracefully, then force
	 * them into reset.  once active.rebooting is set, any or all
	 * of the other cpus may be idling but not servicing interrupts.
	 */
	lock(&active);
	active.rebooting = 1;		/* request other cpus shutdown */
	unlock(&active);
	shutdown(Shutreboot);

	delay(100);	/* let other cpus shut down gracefully */
}

static void
doreset(void)
{
	pcireset();		/* disable pci bus masters & intrs */
	spllo();
	archreset();		/* we can now use the uniprocessor reset */
}

/*
 * shutdown this kernel, jump to trampoline code in Sys, which copies
 * the next kernel (size @ code) into the addresses it was linked for,
 * and jumps to the new kernel's entry address.
 */
void
reboot(void *phyentry, void *code, long size)
{
	void (*tramp)(Reboot *);

	/* there's no config and nowhere to store it */
	drainuart();

	/*
	 * interrupts (including uart) may be routed to any or all cpus, so
	 * shutdown devices, other cpus, and interrupts (rely upon iprint
	 * hereafter).
	 */
	devtabshutdown();
	drainuart();		/* before stopping cpus & interrupts */

	/* other cpus may be idling; put them into reset */
	if (sys->nonline > 1)
		shutothercpus();		/* calls shutdown */

	/*
	 * should be the only processor scheduling now.
	 * any intrs to other cpus will not be delivered hereafter.
	 */
	memset(active.machsmap, 0, sizeof active.machsmap);
	splhi();

	/* we've been asked to just `halt'? */
	if (phyentry == 0 && code == 0 && size == 0) {
		doreset();
		notreached();
	}

	/* setup reboot trampoline function */
	tramp = (void *)PADDR(sys->reboottramp);
	memmove(tramp, rebootcode, sizeof(rebootcode));
	cacheflush();

	spllo();
	iprint("starting new kernel at %#p, via trampoline at %#p...",
		phyentry, tramp);
	delay(1000);
	splhi();
	pcireset();			/* disable pci bus masters & intrs */

	/* off we go - never to return */
	sys->phyentry = (uintptr)phyentry;
	sys->phycode = PADDR(code);
	sys->rebootsize = size;
	ecall();		/* env call from super to machine mode */
	notreached();	
}
