/*
 * Memory and machine-specific definitions.  Used in C and assembler.
 */

#define TOS(stktop)	((stktop)-sizeof(Tos))

#define PAUart0	0x10010000	/* tinyemu only */
#define PAClint	0x02000000	/* appears to be common, if not standard. */

#define KB		1024			/* 2^10 0x0000000000000400 */
#define MB		1048576			/* 2^20 0x0000000000100000 */
#define GB		1073741824		/* 2^30 000000000040000000 */
#define TB		1099511627776ll		/* 2^40 0x0000010000000000 */
#define PB		1125899906842624ll	/* 2^50 0x0004000000000000 */
#define EB		1152921504606846976ll	/* 2^60 0x1000000000000000 */

/* old names */
#define HOWMANY(x, y)	(((x)+((y)-1))/(y))
#define ROUNDUP(x, y)	(HOWMANY((x), (y))*(y))
#define ROUNDDN(x, y)	(((x)/(y))*(y))

#define MIN(a, b)	((a) < (b)? (a): (b))
#define MAX(a, b)	((a) > (b)? (a): (b))

/*
 * Sizes
 */
#define BI2BY		8		/* bits per byte */
#define BY2V		8		/* bytes per double word */
#define BLOCKALIGN	8		/* makes alloc'ns suitable as stacks */
#define FPalign		16		/* required for FXSAVE */
/* old names */
#define BY2WD		4
#define BI2WD		(BI2BY*BY2WD)
#define BY2PG		PGSZ
#define PGROUND(s)	ROUNDUP(s, PGSZ)

#define PGSZ		(4*KB)			/* smallest page size */
#define PGSHFT		12			/* log(PGSZ) */
#define SUPPGSZ		(4*MB)	/* next page size in Sv32; 2^21 otherwise */
#define SUPPGSHFT	22			/* log(SUPPGSZ) */
#define PTSZ		PGSZ			/* page table page size */
/* log(ptes per PTSZ), also virt address bits per PT level */
#define PTSHFT		10
#define PTESHFT		10			/* bits right of PPN in PTE */

#define MACHSZ		PGSZ			/* Mach size, excluding stack */
#define MACHSTKSZ	(6*PGSZ)		/* Mach stack size */
/*
 * max. # of cpus (less than 256 [apics], was 32); costs 1 byte/cpu in
 * sizeof(Page)
 */
#ifdef	TINYEMU
#define	MACHINE_MODE
#define MACHMAX		1
#else
#define MACHMAX		16
#endif

#define KSTACK		(16*1024)		/* Size of Proc kernel stack */
/* ic keeps stack vlong-aligned */
#define STKALIGNSZ	sizeof(vlong)
#define STACKALIGN(sp)	((sp) & ~(STKALIGNSZ-1)) /* bug: assure with alloc */
/*
 * calling a system call stub pushes the arguments onto the user's stack,
 * then the call instruction to the stub pushes a return PC;
 * the system call trap pushes a Ureg at top of kstack, then a PC below it.
 * this skips past the PC to produce the address of the arguments or Ureg,
 * depending upon the stack that sp points into.
 */
#define SKIPSTKPC(sp)	((uintptr)(sp) + sizeof(uintptr))

/*
 * Time
 */
#define HZ		(100)			/* clock frequency */
#define MS2HZ		(1000/HZ)		/* millisec per clock tick */
#define TK2SEC(t)	((t)/HZ)		/* ticks to seconds */

/*
 *  Address spaces
 *
 *  RV32 implements 32 bits of virtual addresses.
 *
 *  Kernel gets loaded at 2GB (KZERO, start of ram);
 *  3GB-1MB is used to hold the Sys and Syscomm data structures.
 *
 *  User is at low addresses; kernel vm starts at KZERO;
 *  KSEG0 maps the first kernmem bytes, one to one (i.e., KZERO -> KZERO),
 *  for the kernel's use, rounded up to the nearest GB;
 *  KSEG2 maps all remaining physical memory, and may contain unmapped gaps,
 *  for device registers or magic ACPI memory.
 *
 *  Separating KSEG0 and KSEG2 (rather than just having KSEG0 cover everything)
 *  seems to be intended to simplify start-up and cover VMAP.
 *
 *  See mmu.c and asm.c.
 */
#define	UZERO		0			/* user segment */
#define	UTZERO		(UZERO+PGSZ)		/* first address in user text */
#define UTROUND(t)	ROUNDUP((t), PGSZ)
#define USTKTOP		(2ul*GB - PGSZ)
#define USTKSIZE	(16*MB)			/* max. size of user stack */
#define TSTKTOP		(USTKTOP-USTKSIZE)	/* end of new stack in sysexec */

/* verify assertion (cond) at compile time */
#define CTASSERT(cond, name) struct { char name[(cond)? 1: -1]; }

#define VZERO		0ul
#define VMAPSZ		(128*MB)		/* multiple of SUPPGSZ */

#define KSEG0		0x80000000u
#define KSEG2		(KSEG0 + physsize)

/* KTZERO is somewhat arbitrary; microchip puts linux kernel at 0x84000000 */
#define KZERO		KSEG0
#define KTZERO		0x80000000		/* see mkfile il -T */

#define isuseraddr(ptr)	((uintptr)(ptr) < KSEG0)

/*
 *  virtual MMU
 */
#define	PTEMAPMEM	(4*MB)		/* memory mapped by a Pte map */
#define	PTEPERTAB	(PTEMAPMEM/PGSZ) /* pages per Pte map */
/* maximum segment map size in Ptes. */
#define SEGMAPSIZE	2048	/* max. MB per segment; was 1984 on 386 */
#define SSEGMAPSIZE	16

/*
 * This is the interface between fixfault and mmuput.
 * Should be in port.  But see riscv.h Pte*.
 */
#define PTEVALID	(1<<0)
#define PTERONLY	(1<<1)
#define PTEWRITE	(1<<2)
#define PTEUSER		(1<<4)		/* unused */
#define PTEUNCACHED	0		/* no such bit on riscv */

#define getpgcolor(a)	0

/*
 * Hierarchical Page Tables.
 * For example, traditional IA-32 paging structures have 2 levels,
 * level 1 is the PD, and level 0 the PT pages; with IA-32e paging,
 * level 3 is the PTROOT, level 2 the PDP, level 1 the PD,
 * and level 0 the PT pages. The PTLX macro gives an index into the
 * page-table page at level 'l' for the virtual address 'v'.
 */
#define NPGSZ 2  /* max. # of page sizes & page table levels, for 32 bits */
#define PTLX(v, l) \
	(uintptr)(((uvlong)(v) >> (((l)*PTSHFT)+PGSHFT)) & ((1<<PTSHFT)-1))
#define PGLSZ(l)	(1ll << (((l)*PTSHFT)+PGSHFT))

/* this can go when the arguments to mmuput change */
#define PPN(x)		((x) & ~(PGSZ-1))		/* GAK */
#define PTEPPN(x)	(((uintptr)(x) / PGSZ) << PTESHFT)

#define SYSEXTEND	(64*KB)	/* size of Sys->sysextend */
#define KMESGSIZE	(16*KB)

/* pages allocated by vmap, before malloc is ready */
#define EARLYPAGES	32

/* offsets in Sys struct, from dat.h */
#define LOWMACHSTK	0
#define LOWMACH		(LOWMACHSTK + MACHSTKSZ + PTSZ + 32*PTSZ)
/* rest are on cpu0 only */
#define LOW0SYS		(LOWMACH + MACHSZ)
#define LOW0SYSPAGE	(LOW0SYS + PGSZ)
#define LOW0MACHPTR	(LOW0SYSPAGE + PGSZ)
#define LOW0END		(LOW0MACHPTR + PGSZ + 2*PTSZ + SYSEXTEND)
#define LOW0ZEROEND	(LOW0END - KMESGSIZE)	/* stop zeroing before Kmesg */

#define HTIFREAD	0x01000000	/* dev 1 cmd 0 */
#define HTIFPRINT	0x01010000	/* dev 1 cmd 1 */
