/*
 * RV32G memory management for 32-bit virtual addresses.
 * specifically, virtual to physical address mappings.
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"

#include "riscv.h"

#define ALIGNED(p, a)	(!(((uintptr)(p)) & ((a)-1)))
#define PADDRINPTE(pte)		(((uintptr)(pte) >> PTESHFT) * PGSZ)
#define PAGEPRESENT(pte)	((pte) & PteP)

/* see mem.h */
/* page table levels start at the leaves and increase up to the root */
#define PDX(v)		PTLX((v), 1)
#define PTX(v)		PTLX((v), 0)

enum {
	VMAP =		VZERO - VMAPSZ,		/* general mappings */
};

static Lock vmaplock;
static Page mach0ptrootpg;

/*
 * max kernel memory use within KSEG0, multiple of 2 or 4 MB.
 * used in mmu.c and map.c.
 */
uintptr kernmem = GB;		/* stay below VMAP */

PTE	asmwalkalloc(usize size);

uintptr
topram(void)
{
	/*
	 * leave room for Sys at top; exclude last MB from KSEG2 mappings.
	 */
	return MB;
}

void
mmuwrpte(PTE *ptep, PTE pte)
{
	Mpl pl;

	pl = splhi();
	wbinvd();
	*ptep = pte;
	wbinvd();
	splx(pl);
}

/* insert a PTE entry for va->pa mapping in NPGSZ-1 level pt. */
void
mappage(Page *ptroot, PTE *pte, uintptr va, uintptr pa)
{
	uint ptebits;

	ptebits = PteR | PteW | PteX | PtePS | PteP | Pteleafmand;
	if (va >= KSEG0)
		ptebits |= PteG;
	else
		ptroot->daddr++;
	if (*pte)
		panic("mappage: pte for va %#p in use", va);
	*pte = PTEPPN(pa) | ptebits;		/* super page leaf */
}

void
idmappage(Page *ptroot, PTE *pte, uintptr va)
{
	mappage(ptroot, pte, va, va);
}

/*
 * set up id map in lower addresses using only top-level
 * (NPGSZ-1) superpages.  called before reboot.
 */
void
mmulowidmap(void)
{
	uintptr pg, pgsz;
	Page *page;
	PTE *pte;

	wbinvd();
	page = m->ptroot;
	/* zero user mapping */
	memset((void *)page->va, 0, sizeof(PTE) * page->daddr);
	page->daddr = 0;

	pgsz = PGLSZ(m->npgsz - 1);	/* on Sv32, 4MB */

	/* populate top-level identity page table for low addresses */
	pte = (PTE *)page->va;
	for (pg = 0; pg < HOWMANY(4ull*GB, pgsz); pg++)
		idmappage(page, pte++, pg * pgsz);
	wbinvd();
}

/* called before mmuinit */
void
pageidmap(void)
{
	uint ptebits;
	uintptr i, va;
	Page *page;
	PTE *pte;

	memset(&mach0ptrootpg, 0, sizeof mach0ptrootpg);
	memset(sys->pteroot, 0, PTSZ/sizeof(PTE));
	m->ptroot = page = &mach0ptrootpg;
	page->pa = page->va = (uintptr)sys->pteroot;
	page->daddr = 0;

	/* populate top-level identity page table */
	pte = (PTE *)page->va;
	for (i = 0; i < VMAP/PGLSZ(1); i++) {
		ptebits = PteR | PteW | PteX | PtePS | PteP | Pteleafmand;
		va = i * PGLSZ(1);
		if (va >= KSEG0)
			ptebits |= PteG;
		*pte++ = PTEPPN(va) | ptebits;		/* super page leaf */
		if (va < KSEG0)
			page->daddr++;
	}
	print("mmu...");
/**/	putsatp(Pagingon | ((uintptr)page->pa / PGSZ));
}

void
mmuflushtlb(Page *)
{
	Mpl s;
	Page *ptroot;

	s = splhi();
	/* zero the top-level user-space PTEs */
	ptroot = m->ptroot;
	if(ptroot->daddr){
		wbinvd();
		memset((char *)ptroot->va, 0, ptroot->daddr*sizeof(PTE));
		wbinvd();
		ptroot->daddr = 0;
	}
	splx(s);
}

void
mmuflush(void)
{
	Mpl s;

	s = splhi();
	up->newtlb = 1;
	mmuswitch(up);
	splx(s);
}

static void
mmuptpfree(Proc* proc, int release)
{
	int l;
	PTE *pte;
	Page **last, *page;

	/*
	 * To do here:
	 *	coalesce the clean and release functionality
	 *	(it's either one or the other, and no need for
	 *	wakeup in mmurelease as not using the palloc pool);
	 *	0-based levels, not 1-based, for consistency;
	 *	fix memset level for PGLSZ(1) super pages;
	 *	use a dedicated datastructure rather than Page?
	 */
	for(l = 1; l < NPGSZ; l++){		/* from leaf-1 to root */
		last = &proc->mmuptp[l];
		if(*last == nil)
			continue;
		for(page = *last; page != nil; page = page->next){
			if(!release){
				wbinvd();
				if(l == 1)		/* leaf-1? */
					memset((char *)page->va, 0, PTSZ);
				pte = (PTE *)page->prev->va;
				mmuwrpte(&pte[page->daddr], 0);
			}
			last = &page->next;
		}
		*last = proc->mmuptp[0];
		proc->mmuptp[0] = proc->mmuptp[l];
		proc->mmuptp[l] = nil;
	}
	m->ptroot->daddr = 0;
}

static Page*
mmuptpalloc(void)
{
	void* va;
	Page *page;

	/*
	 * Do not really need a whole Page structure,
	 * but it makes testing this out a lot easier.
	 * Could keep a cache and free excess.
	 */
	if((page = malloc(sizeof(Page))) == nil){
		print("mmuptpalloc Page\n");
		return nil;
	}
	if((va = mallocalign(PTSZ, PTSZ, 0, 0)) == nil){
		print("mmuptpalloc va\n");
		free(page);
		return nil;
	}

	page->va = PTR2UINT(va);
	page->pa = PADDR(va);
	page->ref = 1;
	return page;
}

/*
 * switch user mmu mappings to proc's.
 */
void
mmuswitch(Proc* proc)
{
	PTE *pte;
	Page *page, *ptroot;

	if(proc->newtlb){
		mmuptpfree(proc, 0);
		proc->newtlb = 0;
	}

	/* zero the top-level user-space PTEs */
	ptroot = m->ptroot;
	if(ptroot->daddr){
		wbinvd();
		memset((char *)ptroot->va, 0, ptroot->daddr*sizeof(PTE));
		wbinvd();
		ptroot->daddr = 0;
	}

	/* populate the top-level PTEs from proc->mmuptp[NPGSZ-1] */
	pte = (PTE *)ptroot->va;
	for(page = proc->mmuptp[NPGSZ - 1]; page != nil; page = page->next){
		/* next level ptr: no leaf bits */
		mmuwrpte(&pte[page->daddr], PTEPPN(page->pa) | PteP);
		if(page->daddr >= ptroot->daddr)
			ptroot->daddr = page->daddr+1;
		page->prev = ptroot;
	}
	wbinvd();
//	idlewake();			/* other runnable procs? */
}

void
mmurelease(Proc* proc)
{
	Page *page, *next;

	/*
	 * See comments in mmuptpfree above.
	 */
	mmuptpfree(proc, 1);

	for(page = proc->mmuptp[0]; page != nil; page = next){
		next = page->next;
		if(--page->ref)
			panic("mmurelease: page->ref %d", page->ref);
		free((void *)page->va);
		free(page);
	}
	if(proc->mmuptp[0] && palloc.r.p)
		wakeup(&palloc.r);
	proc->mmuptp[0] = nil;
	wbinvd();
}

enum {
	Debug = 0,
};

/* returns a physical address from pte */
static PTE *
nextpte(PTE pte)
{
	if (PAGEPRESENT(pte))
		return (PTE *)PADDRINPTE(pte);
	else
		return nil;
}

/*
 * returns va of page table that maps `va' at level.
 * called from mmuput and mmuwalk.
 *
 * there's a trick, called `virtual page table', used on 386 and amd64
 * systems to insert the address of the root page table into the root
 * page table, thus allowing a virtual view of the page table tree as a
 * contiguous array with holes for missing page tables.  this is possible
 * because one can't tell (ignoring the Page Size bit) from looking at a
 * 386/amd64 PTE what level of the tree it corresponds to.  on risc-v,
 * each PTE identifies itself as leaf or interior node, so the trick
 * doesn't seem to be possible.
 */
static PTE*
mmuptpget(uintptr va, int level, int *lvlgotp)
{
	int l, idx, lvlgot;
	PTE pte;
	PTE *ptp, *ptnxt, *kptp;

	if (level < 0 || level >= NPGSZ)
		panic("mmuptpget level %d", level);

	ptp = (PTE *)m->ptroot->pa;		/* level NPGSZ-1 */
	if (ptp == nil)
		panic("mmuptpget: no pt root");
	l = NPGSZ - 1;
	lvlgot = l;
	idx = PTLX(va, l);
	kptp = KADDR((uintptr)ptp);
	pte = kptp[idx];			/* root page pte */
	ptnxt = nextpte(pte);			/* physical */
	if (Debug)
		iprint("mmuptpget 1: va %#p lvl %d pt is %#p: %#p\n",
			va, l, ptp, pte);

	/*
	 * walk toward a leaf PTE from the level-after-root until `level',
	 * return the containing page.
	 */
	while (--l >= level){
		if (!PAGEPRESENT(pte))		/* parent hole? */
			break;
		if (pte & PteLeaf)		/* not a next-level pointer? */
			break;
		lvlgot = l;

		ptp = ptnxt;			/* advance to level l */
		if (l == 0)			/* level 0 has to be a leaf */
			break;
		if (ptp == nil)
			panic("mmuptpget: nil ptp at level %d for %#p", l, va);
		idx = PTLX(va, l);
		kptp = KADDR((uintptr)ptp);
		pte = kptp[idx];		/* pte at next level */
		ptnxt = nextpte(pte);		/* physical */
		if (Debug)
			iprint("mmuptpget 2: va %#p lvl %d pt is %#p: %#p\n",
				va, l, ptp, pte);
	}
	if (Debug)
		iprint("* mmuptpget return va %#p lvl %d -> %#p\n", va, l, ptp);
	if (lvlgotp)
		*lvlgotp = lvlgot;
	if (ptp)
		return KADDR((uintptr)ptp);
	return nil;
}

/* only called from fault.c.  pa may contain PTE attributes in low bits. */
void
mmuput(uintptr va, uintmem pa, Page*)
{
	Mpl pl;
	int l, lvlgot, x, attrs;
	PTE oldpte, ptebits;
	PTE *pte, *ptp;
	Page *page, *prev;

	pte = nil;
	pl = splhi();
	prev = m->ptroot;
	attrs = pa & PteAttrs;
	pa &= ~MASK(PGSHFT);
	oldpte = 0;

	/* supply any missing intermediate page table levels */
	for(l = NPGSZ-1; l >= 0; l--){
		ptp = mmuptpget(va, l, &lvlgot);
		if (0 && ptp == nil)
			iprint("mmuput va %#p pa %#p: ptp %#p level %d\n",
				va, pa, ptp, l);
		x = PTLX(va, l);
		pte = &ptp[x];
		for(page = up->mmuptp[l]; page != nil; page = page->next)
			if(page->prev == prev && page->daddr == x)
				break;
		if(page == nil){		/* no such next-level page? */
			/* allocate a new next-level page table page */
			if(up->mmuptp[0] == 0)
				page = mmuptpalloc();
			else {
				page = up->mmuptp[0];
				up->mmuptp[0] = page->next;
			}
			page->daddr = x;
			page->next = up->mmuptp[l];
			up->mmuptp[l] = page;
			page->prev = prev;

			oldpte = *pte;
			/* next level ptr: no leaf bits */
			mmuwrpte(pte, PTEPPN(page->pa) | PteP);

//			iprint("l %d mmuput: va %#p pte %#p (pa %#p) -> %#p (was %#p)\n",
//				l, page->va, *pte, PADDRINPTE(*pte), pte, oldpte);
			USED(oldpte);
			if(l == NPGSZ-1 && x >= m->ptroot->daddr)
				m->ptroot->daddr = x+1;
		}
		prev = page;
	}

	/* overwrite pte written in last iteration of above loop */
	if (oldpte == 0)
		oldpte = *pte;
	ptebits = PTEPPN(pa)|PteU|PteR|PteX|attrs|PteP|Pteleafmand; /* leaf */
	if(attrs & PteW)
		ptebits &= ~PteX;
//	else
//		ptebits &= ~PteD;
//	ptebits &= ~(PteA | PteD);
	mmuwrpte(pte, ptebits);

//	iprint("mmuput va %#p pa %#p: %#p (pa %#p) -> %#p (was %#p)\n",
//		va, pa, *pte, PADDRINPTE(*pte), pte, oldpte);
	if(*pte == oldpte) {
		iprint("mmuput va %#p pa %#p: %#p (pte pa %#p) -> %#p (was %#p)\n",
			va, pa, *pte, (uintptr)PADDRINPTE(*pte), pte, oldpte);
		panic("mmuput: rewriting pte @ %#p w same contents %#p; fixfault stuck",
			pte, *pte);
	}
	splx(pl);

	invlpg(va);			/* only if old entry valid? */
}

typedef struct {
	char	name;
	int	bit;
} Bitname;
Bitname bitnames[] = {
	'r',	PteR,
	'w',	PteW,
	'x',	PteX,
//	'v',	PteP,
	'd',	PteD,
	'a',	PteA,
	'u',	PteU,
	'g',	PteG,
};

void
indent(int depth)
{
	while (depth-- > 0)
		print("\t");
}

/* print cpu's page table */
void
mmudumplvl(uintptr ptroot, uintptr base, int lvl)
{
	int lastcont;
	vlong i;
	uintptr va;
	Bitname *bp;
	PTE ent, lastpte;
	PTE *pte;

	if (lvl < 0)
		return;
	pte = (PTE *)ptroot;
	indent(NPGSZ - 1 - lvl);
	print("page table level %d at %#p:", lvl, pte);
	lastpte = -1;
	lastcont = 0;
	for (i = 0; i < PGSZ/sizeof(PTE); i++, lastpte = ent) {
		ent = pte[i];
		if ((ent & MASK(PTESHFT)) == (lastpte & MASK(PTESHFT)) &&
		    PADDRINPTE(ent) > PADDRINPTE(lastpte)) {
			/* don't print continuations */
			if (!lastcont) {
				print(" ...");
				lastcont = 1;
			}
			continue;
		}
		lastcont = 0;
		if (!PAGEPRESENT(ent))
			continue;

		print("\n");

		indent(NPGSZ - 1 - lvl);
		va = base + i*PGLSZ(lvl);
		print("pte @ va %#10p idx %3d for va %#10p -> pa %#10p ",
			&pte[i], (int)i, va, (uintptr)PADDRINPTE(ent));
		for (bp = bitnames; bp < bitnames + nelem(bitnames); bp++)
			print("%c", ent & bp->bit? bp->name: '-');
		print(" ");
		if (ent & PteLeaf)
			print("%spage leaf", ent & PtePS? "super-": "");
		else {
			print("next level ptr:\n");
			mmudumplvl((uintptr)KADDR(PADDRINPTE(ent)), va, lvl - 1);

			indent(NPGSZ - 1 - lvl);
			print("----\n");
			lastcont = 0;
		}
		if (PADDRINPTE(ent) & m->pgszmask[lvl] != 0)
			print(", not aligned on (super) page boundary");
	}
	print("\n");
}

void
mmudump(uintptr ptroot)
{
	mmudumplvl(ptroot, 0, NPGSZ - 1);
	print("\n");
}

static void*
earlypagealloc(void)
{
	uintptr pa;
	void *alloc;

	if (sys->freepage == nil)
		sys->freepage = (PTE *)sys->pts;
	if (sys->freepage >= (PTE*)&sys->pts[sizeof sys->pts]) {
		pa = asmwalkalloc(PGSZ);
		return KADDR(pa);
	}
	alloc = sys->freepage;
	sys->freepage += PTSZ/sizeof(PTE);
	return alloc;
}

/*
 * Need a PTSZ physical allocator here.
 * Because space will never be given back
 * (see vunmap below), just malloc it so
 * Ron can prove a point.
alloc = pmalloc(PTSZ)|PteRW|PteP;
 */
PTE *
mmunewpage(PTE *pde)
{
	void *alloc;

	if (mallocok)
		alloc = mallocalign(PTSZ, PTSZ, 0, 0);	/* replace this */
	else
		alloc = earlypagealloc();
	if(alloc == nil)
		panic("pdmap: alloc failed");
	memset(alloc, 0, PTSZ);
	/* no leaf bits: next level ptr to new l0 page (alloc) */
	mmuwrpte(pde, PTEPPN(PADDR(alloc))|PteG|PteP);
	return alloc;
}

/*
 * Add kernel mappings for va -> pa for a section of size bytes.
 * Called only after the va range is known to be unoccupied.
 * May allocate page table pages.
 */
static int
pdmap(uintmem pa, int attr, uintptr va, uintptr size)
{
	uintmem pae;
	PTE *pd, *pde, *pt, *pte, *nxtpt;
	int pdx, pgsz;

	pd = (PTE*)m->ptroot->va;
	if (attr == 0)
		attr = PteR|PteW;
//	print("pdmap: %#p %ld -> %#p\n", va, size, pa);
	for(pae = pa + size; pa < pae; pa += pgsz, va += pgsz){
		pdx = PDX(va);
		pde = &pd[pdx];

		/*
		 * Check if it can be mapped using a big page,
		 * i.e. is big enough and starts on a suitable boundary.
		 * Assume processor can do it.
		 */
		if(ALIGNED(pa, PGLSZ(1)) && ALIGNED(va, PGLSZ(1)) &&
		    pae - pa >= PGLSZ(1)){
			pgsz = PGLSZ(1);
			if(*pde != 0) {	/* l1 slot already allocated? */
				panic("pdmap: *pde %#p != 0 for va %#p lvl 1 -> pa %#p",
					*pde, va, pa);
			}
			/* super page leaf */
			mmuwrpte(pde, PTEPPN(pa)|attr|PtePS|PteG|PteP|Pteleafmand);
		} else{
			if(*pde == 0) {		/* l1 slot unpopulated? */
				nxtpt = mmunewpage(pde);
				if(*pde == 0)
					panic("pdmap: *pde %#p == 0 for va %#p "
						"lvl 0 -> pa %#p", *pde, va, pa);
			} else
				nxtpt = KADDR(PADDRINPTE(*pde));
			USED(nxtpt);

//			pt = &pd[pdx];
			pt = KADDR(PADDRINPTE(pd[pdx]));
			assert(pt != nil);
			pte = &pt[PTX(va)];
			/* normal leaf */
			mmuwrpte(pte, PTEPPN(pa)|attr|PteG|PteP|Pteleafmand);
			pgsz = PGLSZ(0);
		}
	}

	return 0;
}

/* find `count' consecutive empty PTEs in a, there are `n' PTEs to examine */
static int
findhole(PTE* a, int n, int count)
{
	int have, i;

	have = 0;
	for(i = 0; i < n; i++){
		if(a[i] == 0)
			have++;
		else
			have = 0;
		if(have >= count)
			return i+1 - have;
	}
	return -1;
}

enum {
	Pdsz = VMAPSZ/PGLSZ(1),
	Ptsz = PGLSZ(0)/sizeof(PTE),
};

/*
 * Look for free space in the vmap.
 */
static uintptr
vmapalloc(uintptr size)
{
	int i, n, o;
	PTE pte;
	PTE *pd, *pt;

	pd = (PTE *)m->ptroot->va + PDX(VMAP);

	/*
	 * Look directly in the PD entries if the size is
	 * larger than the range mapped by a single entry.
	 */
	if(size >= PGLSZ(1)){
		n = HOWMANY(size, PGLSZ(1));
		if((o = findhole(pd, Pdsz, n)) != -1)
			return VMAP + o*PGLSZ(1);
		return 0;
	}

	/*
	 * Size is smaller than that mapped by a single PD entry.
	 * Look for an already mapped PT page that has room.
	 */
	n = HOWMANY(size, PGLSZ(0));
	for(i = 0; i < Pdsz; i++){
		pte = pd[i];
		if(PAGEPRESENT(pte) && !(pte & PteLeaf)) {
			pt = (PTE *)KADDR(PADDRINPTE(pte));
			if((o = findhole(pt, Ptsz, n)) != -1)
				return VMAP + i*PGLSZ(1) + o*PGLSZ(0);
		}
	}

	/*
	 * Nothing suitable, start using a new PD entry.
	 */
	if((o = findhole(pd, Pdsz, 1)) != -1)
		return VMAP + o*PGLSZ(1);
	return 0;
}

/* return uncached mapping for pa of size */
void*
vmap(uintmem pa, uintptr size)
{
	uintptr va, o, sz;

	DBG("vmap(%#P, %llud)\n", pa, (uvlong)size);

	if(m->machno != 0)
		panic("vmap");

	/*
	 * This is incomplete; the checks are not comprehensive enough.
	 *
	 * To do this properly will require keeping track of the
	 * mappings; perhaps something like kmap, but kmap probably
	 * can't be used early enough for some of the uses.
	 */

	/*
	 * Might be asking for less than a page.
	 * This should have a smaller granularity if
	 * the page size is large.
	 */
	o = pa & ((1<<PGSHFT)-1);
	pa -= o;
	sz = ROUNDUP(size+o, PGSZ);

	if(pa == 0){
		DBG("vmap(0, %llud) pc=%#p\n", (uvlong)size, getcallerpc(&pa));
		return nil;
	}
	ilock(&vmaplock);
	/* we hope that caching has been suitably arranged by hardware */
	if((va = vmapalloc(sz)) == 0 || pdmap(pa, PteR|PteW, va, sz) < 0){
		iunlock(&vmaplock);
		DBG("vmap(%#P, %llud) failed\n", pa+o, (uvlong)size);
		return nil;
	}
	iunlock(&vmaplock);
	DBG("vmap(%#P, %llud) => va %#P\n", pa+o, (uvlong)size, va+o);
//	mmuphysaddr(va + o);		/* apparently not needed */
	return (void *)(va + o);
}

void
vunmap(void *v, uintptr size)
{
	DBG("vunmap(%#p, %llud)\n", v, (uvlong)size);

	if(m->machno != 0)
		panic("vunmap");

	/*
	 * See the comments above in vmap.
	 */

	/*
	 * Here will have to deal with releasing any
	 * resources used for the allocation (e.g. page table
	 * pages).
	 */
	DBG("vunmap(%#p, %llud)\n", v, (uvlong)size);
	USED(v);
}

/*
 * walk page tables for va from root to level.
 * return final PTE* via ret.
 * allocate new page table pages as needed.
 */
int
mmuwalk(uintptr va, int level, PTE** ret, uintptr (*alloc)(usize))
{
	int l, lvlgot;
	Mpl pl;
	uintptr pa;
	PTE *pte, *ptp;

//	DBG("mmuwalk cpu%d: va %#p level %d\n", m->machno, va, level);
	pte = 0;
	pl = splhi();
	for(l = NPGSZ-1; l >= 0; l--){
		ptp = mmuptpget(va, l, &lvlgot);
		if (ptp == 0)
			panic("mmuwalk: no ptp for %#p at lvl %d", va, l);
		pte = &ptp[PTLX(va, l)];
		if(l == level)
			break;

		if (!PAGEPRESENT(*pte)) {	/* no PTE? */
			if(alloc == nil)	/* normal case: no allocator */
				break;
			/* allocate a page for next level pt */
			pa = alloc(PTSZ);
			if(pa == ~(uintptr)0) {
				splx(pl);
				return -1;
			}
			if(pa & (PTSZ-1))
				print("mmuwalk pa %#p unaligned\n", pa); /* DEBUG */
			/* map new next-level page: no leaf bits */
			mmuwrpte(pte, PTEPPN(pa) | PteP);
			if((ptp = mmuptpget(va, l-1, &lvlgot)) == nil)
				panic("mmuwalk: mmuptpget(%#p, %d)", va, l-1);
			wbinvd();
			memset(ptp, 0, PTSZ);
			wbinvd();
		}
		else if(*pte & PtePS)		/* super page leaf? */
			break;
	}
	*ret = pte;
	splx(pl);
	return l;
}

/*
 * Given a VA, find the PA.
 * This is probably not the right interface,
 * but will do as an experiment. Usual
 * question, should va be void* or uintptr?
 */
u64int
mmuphysaddr(uintptr va)
{
	int l;
	PTE *pte;
	uintptr mask, pa;

	pte = 0;
//	if (va >= KZERO && va < KZERO + physsize)	/* ram? */
//		return va - KZERO + physmem;

	l = mmuwalk(va, 0, &pte, nil);
	DBG("mmuphysaddr: va %#p lvl %d pte %#p\n", va, l, pte);
	if(l < 0) {
		DBG("mmuphysaddr: va %#p mapping failed\n", va);
		return ~0ull;
	}

	if (!PAGEPRESENT(*pte)) {
		print("mmuphysaddr: va %#p: pte %#p invalid (%#p)\n",
			va, pte, *pte);
		return ~0ull;
	}
	mask = (1ul << (l*PTSHFT + PGSHFT)) - 1;
	pa = (PADDRINPTE(*pte) & ~mask) + (va & mask);
//	DBG("mmuphysaddr: mask %#p pte %#p *pte %#p va %#p\n", mask, pte, *pte, va);

	DBG("mmuphysaddr: lvl %d va %#p pa %#p\n", l, va, pa);
	return pa;
}

void *
evmap(uintptr pa, uintptr size)
{
	uintptr va;

	va = (uintptr)vmap(pa, size);
	if (va == ~0ull)
		va = pa;			/* and pray */
	/*
	 * don't probe here.  pci, for example, may have to be taken out of
	 * reset first.
	 */
	if  (0 && probeulong((ulong *)va, 0) < 0)
		iprint("evmap: va %#p (for pa %#p) faulted on read\n", va, pa);
	return (void *)va;
}

Clint *
clintaddr(void)
{
	if (m->virtdevaddrs)
		return (Clint *)soc.clint;
	else
		return (Clint *)PAClint;
}

/*
 * map low device registers to >KZERO addresses to keep them
 * visible when user processes are mapped.
 */
static void
vmapdevs(void)
{
	/* these appear to be common, if not standard. */
	soc.clint = vmap(PAClint, 64*KB);

	soc.uart = vmap(uart0regs, PGSZ);
	uart0regs = (uintptr)soc.uart;
	uartsetregs(0, uart0regs);		/* update 8250 struct */

#ifndef TINYEMU
	/* polarfire icicle configuration */
	soc.plic  = vmap(PAPlic, 64*MB);	/* common */
	soc.wdog0 = vmap(PAWdog0, PGSZ);
	soc.rtc   = vmap(PARtc, PGSZ);
	soc.timer = vmap(PATimer, PGSZ);
#else
	/* tinyemu configuration */
//	soc.htif  = vmap(PAHtif, PGSZ);
//	htifpwrregs = (uintptr)soc.htif;
//	htifregs = htifpwrregs;
#endif
	m->virtdevaddrs = 1;
}

void
mmuinitap(void)
{
	uchar *p;
	Page *ptroot;

	/*
	 * GAK: Has to go when each mach is using
	 * its own page table
	 */
	p = (uchar *)m->stack;
	/*
	 * make a copy of the top-level ptroot page,
	 * and point our m->ptroot at it.
	 */
	p += MACHSTKSZ;
	memmove(p, (char *)mach0ptrootpg.va, PTSZ);
	m->ptroot = ptroot = &m->ptrootpriv;
	ptroot->va = PTR2UINT(p);
	ptroot->pa = PADDR(p);
	ptroot->daddr = mach0ptrootpg.daddr;  /* # of user mappings in ptroot */
	if(ptroot->daddr){
		memset(p, 0, ptroot->daddr*sizeof(PTE));
		ptroot->daddr = 0;
	}

	cacheflush();
	DBG("SATP %#p\n", Pagingon | (ptroot->pa / PGSZ));
/**/	putsatp(Pagingon | (ptroot->pa / PGSZ));
	DBG("mach cpu%d: %#p ptroot va %#p -> pa %#p\n", m->machno, m,
		ptroot->va, ptroot->pa);
	m->virtdevaddrs = 1;
}

/*
 * Set up the various kernel memory allocator limits:
 * pmstart/pmend bound the unused physical memory;
 * vmstart/vmend bound the total possible VM used by the kernel;
 * vmunused is the highest virtual address currently mapped
 * and used by the kernel;
 * vmunmapped is the highest virtual address currently mapped
 * by the kernel.
 * Vmunused can be bumped up to vmunmapped before more
 * physical memory needs to be allocated and mapped.
 *
 * This is set up here so meminit can map appropriately.
 */
static void
allocinit(void)
{
	uintptr o, pa, sz, gap;

	o = sys->pmstart;
	gap = o - sys->pmbase;
	sz = ROUNDUP(o, PGLSZ(1)) - o; /* gap from kernel end to next super page */
	pa = asmalloc(0, sz, AsmMEMORY, 0);
	if(pa == 0)
		panic("mmuinit: asmalloc for gap after kernel failed");
	if(pa != o)			/* sanity check */
		panic("mmuinit: pa %#p memstart %#p", pa, o);
	sys->pmstart += sz;
	sys->vmstart = KSEG0;
	sys->vmunused = sys->vmstart + ROUNDUP(gap, PGSZ); /* page after kernel */
	sys->vmunmapped = sys->vmstart + gap + sz;	/* next super page */
	/* round up for meminit on vbox, else it may blow an assertion */
	/* end of kernel data */
	sys->vmend = ROUNDUP(sys->vmstart + kernmem, PGLSZ(1));

	DBG("mmuinit: kseg0: vmstart %#p vmunused %#p\n"
		"\tvmunmapped %#p vmend %#p\n",
		sys->vmstart, sys->vmunused, sys->vmunmapped, sys->vmend);
}

void
mmuinit(void)
{
	int l, lvl;
	PTE *pte;
	Page *page;

	archmmu();
	DBG("mach%d: %#p npgsz %d\n", m->machno, m, m->npgsz);
	if(m->machno != 0) {
		mmuinitap();
		return;
	}

	/* we are cpu0 */
	page = m->ptroot;
	if (0) {				/* already done in pageidmap */
		m->ptroot = page = &mach0ptrootpg;
		page->pa = page->va = (uintptr)sys->pteroot;
	}
	allocinit();
	DBG("mach cpu%d: %#p pt root va %#p -> pa %#p\n", m->machno, m,
		page->va, page->pa);
	USED(page);

	pte = nil;
	for (lvl = NPGSZ-1; lvl >= 0; lvl--)  /* from root to leaves */
		if((l = mmuwalk(KZERO, lvl, &pte, nil)) >= 0) {
			DBG("l %d %#p %#p\n", l, pte, (pte? *pte: 0));
			USED(l);
		}
	mmuphysaddr(PTR2UINT(end));

	vmapdevs();
}
