/*
 * Trap catcher for machine mode of RV32.
 * There is no paging in machine mode, so this handler, at least,
 * must be in identity-mapped memory.
 *
 * Note well that the exact stack layout of a trap is known
 * in various other places too (e.g., main.c, mmu.c, trap.c).
 * The intended layout is: a Ureg struct at the very end of up->kstack if
 * trapped from user mode, or somewhere on m->stack if trapped from kernel,
 * with a pointer to that Ureg just before (below) it, for the
 * argument to trap().
 */

#include "mem.h"
#include "riscvl.h"

#define CALLTRAP \
	MOV	$Mie, R4;\
	CSRRC	CSR(MSTATUS), R4, R0;	/* splhi */\
	MOV	R2, R(ARG);		/* Ureg* */\
	SUB	$(2*XLEN), R2;	/* leave needed room (for return PC & what?) */\
/**/	JAL	LINK, trap(SB);		/* to C: trap(sp) */\
	ADD	$(2*XLEN), R2;\
	MOV	$Mie, R4;\
	CSRRC	CSR(MSTATUS), R4, R0	/* back to splhi */

/* side-effect: saves R4 in regsave, patches Ureg with regsave's R2, 4 */
#define PUSHALLM \
	SAVER1_31;\
	/* patch up Ureg on stack */\
	MOV	SAVEMR2(R(MACH)), R4;\
	MOV	R4, (2*XLEN)(R2);	/* push saved R2 from regsave */\
	/* save CSRs */\
	MOV	CSR(MEPC), R4;		MOV R4, 0(R2);	/* push saved pc */\
	MOV	CSR(MSTATUS), R4;	MOV R4, (32*XLEN)(R2);\
	MOV	CSR(MIE), R4;		MOV R4, (33*XLEN)(R2);\
	MOV	CSR(MCAUSE), R4;	MOV R4, (34*XLEN)(R2);\
	MOV	CSR(MTVAL), R4;		MOV R4, (35*XLEN)(R2);\
	MOV	$Mppmach, R4;		MOV R4, (36*XLEN)(R2)

/* should we restore MIE and MSTATUS? intr enable bits may have been changed */
#define POPCSRM \
	MOV (32*XLEN)(R2), R4; MOV R4, CSR(MSTATUS);\
	MOV (33*XLEN)(R2), R4; MOV R4, CSR(MIE);\
	MOV (34*XLEN)(R2), R4; MOV R4, CSR(MCAUSE);\
	MOV (35*XLEN)(R2), R4; MOV R4, CSR(MTVAL);\
	MOV   0(R2), R4; MOV R4, CSR(MEPC)

/*
 * cope with possibly-nested machine-mode faults, from user or kernel mode.
 * the trap mechanism saved the return address in CSR(MEPC) and masked
 * interrupts.
 */
TEXT mtrap(SB), 1, $-4
	CSRRW	CSR(MSCRATCH), R(MACH), R(MACH)	/* save old m, load new m */

	/* stash a few regs in m->regsave to free working registers */
	MOV	R2, SAVEMR2(R(MACH))	/* save SP */
	MOV	R4, SAVEMR4(R(MACH))	/* save R4 temporary */
	MOV	R6, SAVEMR6(R(MACH))	/* save R6 up */

	MOV	CSR(MCAUSE), R6
	MOV	$Rv32intr, R4
	AND	R4, R6
	BEQ	R6, except
	/*
	 * punt interrupts to supervisor mode.
	 * this should run to MRET without interruption.
	 */
intr:
	MOV	$(Mtie|Meie|Msie), R2
	MOV	CSR(MIP), R4
	AND	R2, R4			/* isolate mach intr bits */

	CSRRC	CSR(MIP), R4, R0	/* clear MIP *ie bits */
	SRL	$2, R4
	CSRRS	CSR(SIP), R4, R0 /* set corr. SIP bits to punt to super */

	MOV	R4, R6
	AND	$Stie, R4
	BEQ	R4, notclock

	/* clock interrupt */
	MOVW	MACHHARTID(R(MACH)), R4
	MOV	$BY2V, R2
	MUL	R2, R4
	MOV	$(PAClint+4*4096), R2	/* phys addr in machine mode */
	ADD	R4, R2			/* &clint->mtimecmp[m->hartid] */
	MOV	$-1, R4
	MOV	R4, 0(R2)
	SRL	$1, R4
	MOV	R4, 4(R2)		/* dismiss timer intr */
notclock:
	MOV	SAVEMR6(R(MACH)), R6	/* R6 up */
	MOV	SAVEMR4(R(MACH)), R4	/* R4 temporary */
	MOV	SAVEMR2(R(MACH)), R2	/* SP */
	CSRRW	CSR(MSCRATCH), R(MACH), R(MACH)	/* restore r7, reload saved m */
	FENCE
	MRET			/* the super intr is awaiting our return */

	/*
	 * exception from kernel (super) or user mode.
	 * going to have to save & restore almost all the registers.
	 * should be rare.  shouldn't get here from user mode.
	 */
except:
	MOV	CSR(MSTATUS), R2
	MOV	$Mpp, R4
	AND	R4, R2
	MOV	SAVEMR4(R(MACH)), R4	/* R4 temporary */
	BEQ	R2, mfromuser

	MOV	SAVEMR6(R(MACH)), R6	/* R6 up */
//	MOV	SAVEMR4(R(MACH)), R4	/* R4 temporary */

	/*
	 * exception from kernel mode: SB, MACH, R2, USER were all okay,
	 * but we stepped on MACH (a no-op) and SP above.
	 */
	MOV	SAVEMR2(R(MACH)), R2
	SUB	$UREGSIZE, R2		/* room for pc, regs, csrs */
	PUSHALLM

	MOV	$setSB(SB), R3
/**/	CALLTRAP

	/* pop regs except MACH, USER, SB, R2; pop CSRs */
	POPCSRM
	POPR8_31
	/*
	 * be sure to pop registers from Ureg, so that changes to this
	 * Ureg elsewhere (e.g. trap) are reflected.
	 * MACH, R2, SB are still correct; USER might not be.
	 */
	MOV	(2*XLEN)(R(MACH)), R(USER)	/*  up = m->proc */
	MOV	(1*XLEN)(R2), R1
	MOV	(5*XLEN)(R2), R5
	MOV	(4*XLEN)(R2), R4	/* restore saved R4 */
	MOV	(2*XLEN)(R2), R2
	MRET				/* to kernel CSR(MEPC) address */

mfromuser:
	/*
	 * trapped from user mode: SB, MACH, R2, USER all needed reloading.
	 * old MACH value is in CSR(MSCRATCH).  R2 is saved in regsave.
	 */
	/* stash a few regs in m->regsave to free working registers */
	MOV	R6, SAVEMR6(R(MACH))	/* save old up */

	MOV	(2*XLEN)(R(MACH)), R(USER)	/*  up = m->proc */
	MOV	(2*XLEN)(R(USER)), R2	/*  up->kstack */
	ADD	$KSTACK, R2
	AND	$~7, R2			/* vlong alignment */
	SUB	$UREGSIZE, R2		/* room for pc, regs, csrs */

	PUSHALLM			/* patches R2, 4 from regsave in Ureg */
	MOV	SAVEMR6(R(MACH)), R3
	MOV	R3, (6*XLEN)(R2)	/* patch old r6 into Ureg */

	MOV	$setSB(SB), R3
/**/	CALLTRAP
	CSRRC	CSR(MSTATUS), $Mie, R0;	/* back to splhi */

	/* pop regs */
	POPCSRM
	POPR8_31

	/*
	 * be sure to pop registers from Ureg, so that changes to this
	 * Ureg elsewhere (e.g. trap) are reflected.
	 */
	MOV	(6*XLEN)(R2), R6
	MOV	(3*XLEN)(R2), R3
	MOV	(1*XLEN)(R2), R1
	MOV	(5*XLEN)(R2), R5
	MOV	(4*XLEN)(R2), R4	/* restore saved R4 */
	MOV	(2*XLEN)(R2), R2
	CSRRW	CSR(MSCRATCH), R(MACH), R(MACH) /* restore old m, reload saved m */
	MRET				/* to CSR(MEPC) address */
