/*
 * interface to sbi firmware
 * TODO: written for rv64 and not yet retrofitted to rv32
 *
 * R10 contains my hart id at start.
 *
 * follows risc-v unix system call abi, which follows the calling
 * conventions of risc-v elf psabi: same as function calls (presumably
 * gcc's) except that the call instruction is ECALL, and a7 (R17) contains the
 * sbi ulong extension id.  R16 may hold a ulong additional function ID.
 * SBI returns error code (`sbiret') in R10, value in R11.
 * error code is success, negatives are failures.
 *
 * what the risc-v abi calls a0 is R10, ... a7 is R17.
 * initial arguments go in a0-a5 (a6 & a7 are reserved by sbi).
 * in rv32, vlongs go in even-odd pairs.
 * sp points at first arg that won't fit in registers.
 * abi requires sp to be 16-byte aligned.
 * we have to enforce that manually; jc won't.
 * r5-7 and r28-31 may be trashed (risc-v c calling standard).
 * jc will have saved r28-31 if they were in use.
 */

/*
 * legacy extensions, function 0:
 * ext 0 set_timer(uvlong);
 * 4	send_ipi(ulong *hartmaskp);
 * 5	remote_fence_i(ulong *hartmaskp);
 * 6	remote_sfence_vma(ulong *hartmaskp, ulong start, ulong size);
 * 7	remote_sfence_vma_asid(ulong *hartmaskp, ulong start, ulong size,
 *		ulong asid);
 * 2	int console_getchar(void);
 * 1	console_putchar(int);
 * 8	shutdown(void);
 *
 * v0.2 extension 0x48534d
 * func 0 hart_start(ulong hartid, ulong resume_addr, ulong resume_new_a1);
 *		at resume, hart will see: R10 is hartid, R11 is resume_new_a1,
 *		satp 0, sie 0
 * 1	hart_stop(void);
 * 2	hart_status(ulong hartid);
 */
#include "mem.h"
#include "riscvl.h"

#define SPLHI	CSRRC	CSR(SSTATUS), $Sie, R0

#define SBIALIGN	16		/* alignment assumed by SBI calls */

#ifndef MACHINE_MODE
#define SBICALL(ext, func, arg1) \
	MOV	R1, R24; \
	MOV	R5, R21;		/* save extern regs */ \
	MOV	R6, R22; \
	MOV	R7, R23; \
	MOV	R2, R20; \
	AND	$(~(SBIALIGN-1)), R2;	/* align stack */ \
	MOV	ext, R17; \
	MOV	func, R16; \
	MOV	arg1, R10; \
	FENCE; \
	ECALL; \
	MOV	R20, R2; \
	MOV	R23, R7; \
	MOV	R22, R6; \
	MOV	R21, R5; \
	MOV	R24, R1

#define LEGSBICALL(ext, arg1) \
	SBICALL(ext, $0, arg1)

TEXT sbiputc(SB), 1, $-4
	LEGSBICALL($1, R8)
	RET
#endif

TEXT setsbitimer(SB), 1, $-4
#ifndef MACHINE_MODE
	LEGSBICALL($0, R8)
#endif
	RET

TEXT sbisendipi(SB), 1, $XLEN		/* vlong sbisendipi(uvlong hartbm[]) */
#ifndef MACHINE_MODE
	LEGSBICALL($4, R8)		/* R8 is virt addr of hartid bitmap */
	MOV	R10, R8			/* return error code */
#else
	MOV	R0, R8
#endif
	RET

TEXT sbiclearipi(SB), 1, $XLEN		/* sbiclearipi(void) */
#ifndef MACHINE_MODE
	LEGSBICALL($3, R0)
#else
	MOV	R0, R8
#endif
	RET

TEXT sbishutdown(SB), 1, $-4
#ifndef MACHINE_MODE
	LEGSBICALL($8, R0)
#else
	MOV	R0, R8
#endif
	RET

/* returns 0 if ext is not supported */
TEXT sbiprobeext(SB), 1, $XLEN		/* sbiprobeext(ulong ext) */
#ifndef MACHINE_MODE
	SBICALL($0x10, $3, R8)
	BLT	R10, noext
	MOV	R11, R10		/* replace error with status */
noext:
	MOV	R10, R8			/* return error code */
#else
	MOV	R0, R8
#endif
	RET

/*
 * v0.2 extensions
 */

// funcs: 0 start, 1 caller returns to SBI, 2 status
TEXT sbihartstart(SB), 1, $XLEN	/* sbihartstart(hartid, phys_start, private) */
#ifndef MACHINE_MODE
	MOV	XLEN(FP), R11		/* phys_start */
	MOV	(2*XLEN)(FP), R12	/* private */
	MOV	R0, R13			/* accident prevention */
	/* args are hartid, phys_start, private; all vlongs */
	SBICALL($HSM, $0, R8)		/* start hart R8 */
	MOV	R10, R8			/* return error code */
#else
	MOV	R0, R8
#endif
	RET

// funcs: 0 start, 1 caller returns to SBI, 2 status
// status results: 0 started, 1 stopped, 2 start pending, 3 stop pending
TEXT sbihartstatus(SB), 1, $XLEN	/* sbihartstatus(hartid) */
#ifndef MACHINE_MODE
	/* args are hartid, phys_start, private; all vlongs */
	SBICALL($HSM, $2, R8)		/* stat hart R8 */
	BLT	R10, stserr
	MOV	R11, R10		/* replace error with status */
stserr:
	MOV	R10, R8			/* return error code */
#else
	MOV	R0, R8
#endif
	RET
