/*
 * initialise cpu enough to call main.
 * assume we are entered in the identity map, compiled for KZERO,
 * and loaded there, such that _main is at KZERO.
 * we hope to be started as RV32, but try to switch to it if not.
 *
 * we must switch from machine mode to supervisor mode in order to get
 * paging active in kernel mode to access user pages,
 * but tinyemu, at least, will only deliver timer interrupts in machine mode.
 *
 * never use TEXT ... $4; use $8 to keep stack vlong-aligned for ic.
 *
 * what if we are started in supervisor mode?  it's hopeless, i think.
 */
#include "mem.h"
#include "riscvl.h"

#undef	DEBUG

#ifdef	DEBUG
/* 0x20 in Lsr (uart reg 5) is Thre (Thr empty) */
#define CONSPUT(c) \
	SAVECONSREGS; \
	MOV	uart0regs(SB), R(UART0);		/* console uart */ \
	MOVW	(5*4)(R(UART0)), R(TMP); \
	AND	$0x20, R(TMP); \
	BEQ	R(TMP), -2(PC); \
	MOV	c, R(TMP); \
	MOVW	R(TMP), (R(UART0)); \
	RESTCONSREGS
#else
#define CONSPUT(r)
#endif

#define RSYS	R11
#define UART0	21
#define RSYSRB	R26
#define CSRZERO	R27
#define MEMBASE R28
#define RTMP	R29
#define TMP	29
#define MACHNO	R30
#define HARTID	R31

#define INITSTKSIZE	(8*KB)

#define ALIGNPAT 0x12345678

TEXT _main(SB), 1, $-4			/* _main avoids libc's main9.s */
	CSRRC	CSR(SSTATUS), $(Mie|Sie), R0	/* splhi */
	MOV	$setSB(SB), R3
	JAL	MEMBASE, 1(PC)		/* used to compute physmem below */

	/* if this traps, we're in supervisor mode and hosed. */
	CSRRC	CSR(MSTATUS), $Mie, R0

	MOV	uart0regs(SB), R(UART0)
	MOV	R0, CSRZERO
	MOV	CSRZERO, CSR(SATP)	/* paging off; rs can't be R0 */
#ifdef MACHINE_MODE
	MOV 	CSR(MHARTID), MACHNO
	MOV	CSR(MHARTID), HARTID

	MOV	$memfault(SB), R9
	MOV	R9, CSR(MTVEC)
	MOV	CSRZERO, CSR(MEDELEG)
#endif
	/*
	 * verify presence of uart registers
	 */
	MOV	$smemfault(SB), R9
	MOV	R9, CSR(STVEC)

	MOV	$PAUart0, R9		/* hardwired until data seg okay */
	MOV	R9, uart0regs(SB)
	MOVW	R0, probebad(SB)
TEXT probe1(SB), 1, $-4
	MOVW	(5*4)(R9), R0		/* read uart status reg, if present */
TEXT probe2(SB), 1, $-4
	FENCE
TEXT probe3(SB), 1, $-4
	MOVW	probebad(SB), R12
	BNE	R12, nouart		/* present? */
	CONSPUT($'u')
	CONSPUT($'0')
//	MOV	R9, phyuart(SB)

	MOV	CSRZERO, CSR(STVEC)

	/* check data segment alignment */
	MOV	$ALIGNPAT, R10
	MOV	dataalign(SB), RTMP
	BNE	R10, RTMP, datasegunaligned
alignok:

	/*
	 * we've established basic sanity.
	 */
	MOV	$MACHMAX, R12
	BGEU	R12, MACHNO, wfi /* not enough data structures for this cpu? */

	/*
	 * set up a temporary stack for this cpu.
	 */
	MOV	$initstks(SB), R2
	MOV	$INITSTKSIZE, R10
	ADD	$1, MACHNO, R12
	MUL	R12, R10
	ADD	R10, R2			/* sp = &initstks[m->machno+1] */
	AND	$~7, R2			/* vlong alignment */

	/*
	 * we are now in a pretty normal C environment
	 * except that BSS has't been zeroed yet.
	 */
	MOV	$4, R9
	SLL	$31, R9
	BEQ	R9, now32		/* shifted off the left end (rv32)? */

	/*
	 * we are (and have been) running as RV64 (or RV128 [unlikely]).
	 * switch to RV32 ISA, since that's what we expect.
	 */
	JAL	LINK, torv32(SB)

	MOV	$4, R9
	SLL	$31, R9
	BNE	R9, notrv32	/* not shifted off the left end (rv64)? */

	/*
	 * we succeeded, but tinyemu hasn't completely made the switch.
	 * now we're running as RV32 in machine mode.
	 * repeat stuff executed as RV64 above, as needed.
	 */
	MOV	$setSB(SB), R3

	MOV	$PAUart0, R9		/* hardwired until data seg okay */
	MOV	R9, uart0regs(SB)
	CONSPUT($'6'); CONSPUT($'4'); CONSPUT($'-'); CONSPUT($'>')

	JAL	MEMBASE, 1(PC)		/* used to compute physmem below */
	MOV 	CSR(MHARTID), MACHNO

	/*
	 * fix hartid<->machno mapping if hart0 is hobbled.
	 */
	MOV	hobbled(SB), R9
	BEQ	R9, nothobbled
	JAL	LINK, hartidcanon
nothobbled:
	/*
	 * for other cpus, wait for the hounds to be released.
	 */
	BEQ	MACHNO, cpu0
	JAL	LINK, awaithounds(SB)	/* put cpu to sleep; sets R(MACH) */
	JMP	machnot0
cpu0:
now32:
	/*
	 * we're on cpu0.  set physmem, physsize, sys
	 */
	MOV	$setSB(SB), R3

	MOV	$PAUart0, R9		/* hardwired until data seg okay */
	JAL	LINK, setsizes(SB)	/* uses R9-R20 */
	JAL	LINK, setsys(SB)	/* also puts sys in RSYS */

	/* we're cpu0, so sys->machptr is not populated yet */
	ADD	$LOWMACH, RSYS, R(MACH)
	JAL	LINK, exportcsrs(SB)
machnot0:
	/*
	 * all cpus eventually end up here.
	 * identify this hart.
	 */
	ADD	$'0', MACHNO, R26
	CONSPUT($'h'); CONSPUT(R26); CONSPUT($' ')

	/*
	 * all cpus eventually end up here, as RV32.
	 */
	MOV	$setSB(SB), R3
	CONSPUT($'R'); CONSPUT($'V'); CONSPUT($'3'); CONSPUT($'2')
	CONSPUT($' ')
	BNE	MACHNO, machnot0

	MOV	$LOW0SYSPAGE, RSYSRB
	ADD	RSYS, RSYSRB
	MOV	$1, RTMP
	MOV	RTMP, RBSTALL(RSYSRB)

	/*
	 * set up for supervisor mode, needed for paging in kernel.
	 * mostly punting faults to supervisor mode, preparing for clock
	 * interrupts, and disabling paging.
	 */
	ADD	$(LOWMACHSTK+MACHSTKSZ-8), RSYS, R2 /* establish stack in sys */
	AND	$~7, R2			/* vlong alignment */
	MOV	R0, R(USER)

	JAL	LINK, disintr(SB)
	JAL	LINK, settraps(SB)	/* needs R(MACH) set */
#ifdef MACHINE_MODE
	JAL	LINK, delegate(SB)
	/* export machine csrs to C */
	MOV	CSR(MIDELEG), R12	/* how much of it stuck? */
	MOV	R12, mideleg(SB)
	JAL	LINK, setmstatus(SB)	/* set [MS]STATUS for switch */
	JAL	LINK, enabmintr(SB)
#endif

	MOV	CSRZERO, CSR(SATP)	/* paging off; rs can't be R0 */

	/* make the switch to super mode */
CONSPUT($'-'); CONSPUT($'>')
	MOV	$super(SB), R12
	MOV	R12, CSR(MEPC)
	FENCE_I
	/* 
	 * Note: in tinyemu, this goes off the rails, iff we started in RV64.
	 * It appears that addresses continue to be sign extended to 64 bits,
	 * but those addresses now cause (endless) exceptions.
	 */
	MRET				/* leap to (MEPC) in super mode */

/*
 * Begin supervisor mode.
 */

TEXT super(SB), 1, $-4
	CONSPUT($'S')

	/*
	 * we are in identity space already on polarfire & tinyemu;
	 * KZERO = physmem.  this mapping may be typical of rv32 systems.
	 *
	 * we would have to build at least top-level page tables
	 * and enable paging here if kernel virt != phys.
	 */
	CSRRC	CSR(SSTATUS), $Sie, R0		/* splhi */
	MOV	$(Sum|Mxr), R12
	CSRRS	CSR(SSTATUS), R12, R0	/* super user access, exec readable */

	MOV	$Seie, R9			/* super global intr enable */
	CSRRS	CSR(SIE), R9, R0

	MOVW	HARTID, MACHHARTID(R(MACH))

	/*
	 * escape to main in C, at last.
	 */
	CONSPUT($'\n')
//	AND	$~7, R2				/* vlong alignment for ic */

	MOV	MACHNO, R(ARG)			/* machno */
/**/	JAL	LINK, main(SB)			/* no return */

	FENCE
wfi:
	WFI
	JMP	wfi

datasegunaligned:
	/* ah, shit; kernel was linked with wrong -R rounding */
	CONSPUT($'?'); CONSPUT($'d'); CONSPUT($'a'); CONSPUT($'t')
	CONSPUT($'a'); CONSPUT($' '); CONSPUT($'s'); CONSPUT($'e')
	CONSPUT($'g'); CONSPUT($' '); CONSPUT($'u'); CONSPUT($'n')
	CONSPUT($'a'); CONSPUT($'l'); CONSPUT($'i'); CONSPUT($'g')
	CONSPUT($'n'); CONSPUT($'e'); CONSPUT($'d'); CONSPUT($'!')
	CONSPUT($'?')
	JMP	alignok

TEXT torv32(SB), 1, $8
	/* we're in RV64 mode */
	/* set super & user xlens (only accessible in mach mode) */
	MOV	$((Mxlen32<<2) | Mxlen32), R18
	SLLI64(32, 18, 18)
	MOV	$(Mxr|Sum|Mppsuper), R19
	OR	R19, R18
	MOV	R18, CSR(MSTATUS)

	MOV	CSRZERO, CSR(SEPC)

	MOV	CSR(MISA), R9
	BEQ	R9, failed			/* can't do it */
	MOV	R0, R10
	SUB	$1, R10
	SRLI64(2, 10, 10)			/* mask except MXL */
	AND	R10, R9				/* zero MXL bits */
	MOV	$Mxlen32, R10
	SLLI64(62, 10, 10)
	OR	R10, R9				/* set RV32 value in MXL */

	FENCE
	FENCE_I
/**/	MOV	R9, CSR(MISA)			/* switch to RV32, we hope */
	FENCE_I

failed:
	RET

nouart:
	CONSPUT($'?');
	CONSPUT($'u');
	JMP	wfi

notrv32:
	/* gah!  failed to switch to RV32. */
	CONSPUT($'s'); CONSPUT($'t'); CONSPUT($'u'); CONSPUT($'c')
	CONSPUT($'k'); CONSPUT($' '); CONSPUT($'i'); CONSPUT($'n')
	CONSPUT($' '); CONSPUT($'R'); CONSPUT($'V'); CONSPUT($'6')
	CONSPUT($'4'); CONSPUT($'!')
	JMP	wfi

hartidcanon:
	/*
	 * swap roles of hart 0 & 1 to be cpus 1 & 0.
	 */
	MOV	$1, RTMP
	BEQ	RTMP, MACHNO, cpu0hart
	BEQ	R0, MACHNO, cpu1hart
	JMP	allharts
cpu0hart:
	MOV	R0, MACHNO
	JMP	allharts
cpu1hart:
	MOV	$1, MACHNO
allharts:
	RET

TEXT memfault(SB), $-4
	MOV	$1, R17
	MOVW	R17, probebad(SB)

	MOV	CSR(MEPC), R17
	ADD	$4, R17  /* advance to next instr; assume aligned by TEXT */
	MOV	R17, CSR(MEPC)
	MRET

TEXT smemfault(SB), $-4
//CONSPUT($'!')
	MOV	$1, R17
	MOVW	R17, probebad(SB)

	/* advance to next instr; assume aligned by TEXT */
	MOV	CSR(SEPC), R17
	ADD	$4, R17
	MOV	R17, CSR(SEPC)
	SRET

TEXT setsizes(SB), 1, $8
	/* if unset, set physmem from MEMBASE */
	/* on polarfire, phys ram = virtual */
	MOV	physmem(SB), RTMP
	BNE	RTMP, memaddrset
	MOV	$~MASK(PGSHFT), RTMP
	AND	MEMBASE, RTMP
	MOV	RTMP, physmem(SB)	/* physical start of memory */
memaddrset:
	/*
	 * if unset in config, set physsize by sizing memory.  it requires
	 * physmem to be set first.
	 */
	MOV	physsize(SB), R12
	BNE	R12, sizeset

	MOV	$memfault(SB), R9
	MOV	R9, CSR(MTVEC)
	MOV	CSRZERO, CSR(MEDELEG)

#define START R9
#define ADDR R14
#define SAVE R18

	MOV	R0, probebad(SB)
	MOV	physmem(SB), START
	MOV	START, ADDR
	MOV	$(-VMAPSZ), R20		/* upper bound of probing */
	MOV	$(64*MB), R21		/* granularity of probing */
	MOV	$-1, R12
loop:
	BGEU	R20, ADDR, sizedone
TEXT alignr<>(SB), 1, $-4		/* simplify memfault */
	MOV	0(ADDR), SAVE		/* save *addr's original contents */
TEXT alignw<>(SB), 1, $-4		/* simplify memfault */
	MOV	ADDR, 0(ADDR)		/* write addr to *addr */
TEXT alignf<>(SB), 1, $-4		/* simplify memfault */
	FENCE
	MOV	probebad(SB), R15
	BNE	R15, sizedone		/* write trapped? */

	MOV	0(ADDR), R13
	BNE	ADDR, R13, sizedone	/* read back differently? */

	XOR	R12, ADDR, RTMP
	MOV	RTMP, 0(ADDR)		/* write ~addr to *addr */
	FENCE
	MOV	0(ADDR), R13
	BNE	RTMP, R13, sizedone	/* read back differently? */

	MOV	SAVE, 0(ADDR)		/* restore *addr to original contents */
	ADD	R21, ADDR		/* advance to next probe address */
	JMP	loop
sizedone:
	SUB	START, ADDR, R13 // R13 = size, ADDR = memend, START = physmem
	MOV	R13, physsize(SB)
sizeset:
	RET

TEXT settraps(SB), 1, $-4
	/* setup usual trap vectors */
#ifdef MACHINE_MODE
	MOV	$mtrap(SB), R9
	MOV	R9, CSR(MTVEC)
	MOV	R(MACH), CSR(MSCRATCH)	/* set for trap use */
#endif
	MOV	$strap(SB), R9
	MOV	R9, CSR(STVEC)
	MOV	R(MACH), CSR(SSCRATCH)	/* set for trap use */
	RET

TEXT setmach(SB), 1, $-4		/* compute mach addr for this cpu */
	MOV	sys(SB), RSYS
	ADD	$LOW0MACHPTR, RSYS, R12	/* sys->machptr */
	MOV	$BY2WD, RTMP
	MUL	MACHNO, RTMP		/* byte offset for hartid */
	ADD	R12, RTMP		/* RTMP = &sys->machptr[hartid] */
	MOV	(RTMP), R(MACH)		/* m = sys->machptr[hartid] */
	RET

TEXT delegate(SB), 1, $-4
	MOV	$-1, R12
	MOV	R12, CSR(MCOUNTEREN)	/* expose timers to super */
	MOV	R12, CSR(MIDELEG)	/* punt intrs to super */
	MOV	$~(1<<Envcallsup), R13
	AND	R13, R12
	/* punt except.s to super but env call from super */
	MOV	R12, CSR(MEDELEG)
	FENCE
	RET

TEXT setsys(SB), 1, $-4			/* set sys */
	MOV	physmem(SB), RSYS
	MOV	physsize(SB), R12
	ADD	R12, RSYS		/* RSYS = phys memory end */
	SUB	$MB, RSYS		/* RSYS = Sys = top of usable ram */
	MOV	RSYS, sys(SB)		/* for C use */
	RET

TEXT exportcsrs(SB), 1, $-4
#ifdef MACHINE_MODE
	/* export csrs to C */
	MOV	CSR(MISA), R9
	MOV	R9, misa(SB)
#ifndef TINYEMU		/* csrs mandatory yet missing in tinyemu */
	MOV	CSR(MVENDORID), R12
	MOV	R12, mvendorid(SB)
	MOV	CSR(MARCHID), R12
	MOV	R12, marchid(SB)
	MOV	CSR(MIMPLID), R12
	MOV	R12, mimplid(SB)
#endif
#endif
	RET

TEXT setmstatus(SB), 1, $-4
	MOV	CSR(MSTATUS), R12
	MOV	$~(Mpp|Ube|Tvm|Tw|Sie|Mie|Spie|Mpie), R13
	AND	R13, R12		/* splhi, le, no traps */
	MOV	$(Mppsuper|Spp), R13
	OR	R13, R12
	MOV	R12, CSR(MSTATUS)	/* set previous mode to super */
#ifndef TINYEMU		/* csrs mandatory yet missing in tinyemu */
	/* gets ill. inst. trap on tinyemu after switching from rv64 */
	MOV	$(Sbe|Mbe), R13
	CSRRC	CSR(MSTATUSH), R13, R0
#endif
	RET

TEXT disintr(SB), 1, $-4	/* disable all interrupts in all modes */
#ifdef MACHINE_MODE
	MOV	$(Msie|Mtie|Meie), R9	/* no machine mode intrs yet */
	CSRRC	CSR(MIE), R9, R0
#endif
	MOV	$(Ssie|Stie|Seie), R9	/* no super mode intrs yet */
	CSRRC	CSR(SIE), R9, R0

	/* disable more interrupts, choose LE, let kernel do what it needs */
	MOV	$(Ube|Sie|Spie), R12
	CSRRC	CSR(SSTATUS), R12, R0	/* splhi, le, prev super */
	MOV	$(Sum|Mxr), R12
	CSRRS	CSR(SSTATUS), R12, R0	/* super user access, exec readable */

	MOV	$(PAClint+4*4096), R9
	MOV	$BY2V, R13
	MUL	R13, HARTID, R12
	ADD	R12, R9			/* &clint->mtimecmp[m->hartid] */
	MOV	$-1, R10
	MOV	R10, 0(R9)
	SRL	$1, R10
	MOV	R10, 4(R9)		/* prevent timer intr */
	FENCE
	RET

TEXT enabmintr(SB), 1, $-4		/* enable a few mach intrs */
CONSPUT($'M')
	MOV	$Mie, R9
	CSRRS	CSR(MSTATUS), R9, R0
	MOV	$Mtie, R9		/* machine timer intr on, not super */
	CSRRS	CSR(MIE), R9, R0
	FENCE
	RET

TEXT awaithounds(SB), 1, $8
	/*
	 * for other cpus, wait for stallothers, then m->online, then continue,
	 * which will ensure isa & switching to supervisor mode, main will call
	 * squidboy, which must copy (or have copied) root pt to a
	 * newly-allocated one.
	 */
#ifdef SMP
stall:
	MOV	stallothers(SB), RTMP
	BNE	RTMP, stall
#endif
	/*
	 * we're not on cpu0, which is now up, thus sys->machptr is populated.
	 * set m (R(MACH)).
	 */
	JAL	LINK, setmach(SB)
wait:
	/* wait for the hounds to be released */
	MOV	ONLINE(R(MACH)), R9	/* m->online */
	BEQ	R9, wait
	RET

	GLOBL	dataalign(SB), $4
	DATA	dataalign(SB)/4, $ALIGNPAT

	GLOBL	initstks(SB), $((MACHMAX+2)*INITSTKSIZE)
