/*
 * Trap catcher for supervisor mode of RV32.
 *
 * The strategy is to exchange R(MACH) and CSR(SSCRATCH), which is pre-loaded
 * with this CPU's Mach pointer, save all registers and a few CSRs on a stack
 * (empty up->kstack if trap came from user mode), call the common trap()
 * function, restore most or all of the stacked registers and CSRs, exchange
 * R(MACH) and CSR(SSCRATCH), and return from trap via SRET to the address in
 * CSR(SEPC).  Depending upon trap()'s action, the called program could
 * potentially continue unaware of the trap.
 *
 * Note well that the exact stack layout of a trap is known in various other
 * places too (e.g., main.c, mmu.c, trap.c).  The intended layout is: a Ureg
 * struct at the very end of up->kstack if trapped from user mode, or
 * somewhere on m->stack if trapped from kernel, with a pointer to that Ureg
 * just before (below) it, for the argument to trap().
 */

#include "mem.h"
#include "riscvl.h"

#define CALLTRAP \
	MOV	R2, R(ARG);		/* Ureg* */\
	SUB	$(2*XLEN), R2;	/* leave needed room (return PC & dummy arg?) */\
/**/	JAL	LINK, trap(SB);		/* to C: trap(sp) */\
	ADD	$(2*XLEN), R2

/* side-effect: saves R4 in regsave, patches Ureg with regsave's R2, 4 */
#define PUSHALLS \
	SAVER1_31;\
	/* patch up Ureg on stack */\
	MOV	SAVESR2(R(MACH)), R4;\
	MOV	R4, (2*XLEN)(R2);	/* push saved R2 from regsave */\
	/* save CSRs */\
	MOV	CSR(SEPC), R4;		MOV R4, 0(R2);	/* push saved pc */\
	MOV	CSR(SSTATUS), R4;	MOV R4, (32*XLEN)(R2);\
	MOV	CSR(SIE), R4;		MOV R4, (33*XLEN)(R2);\
	MOV	CSR(SCAUSE), R4;	MOV R4, (34*XLEN)(R2);\
	MOV	CSR(STVAL), R4;		MOV R4, (35*XLEN)(R2);\
	MOV	$Mppsuper, R4;		MOV R4, (36*XLEN)(R2)

/* should we restore SIE and SSTATUS? intr enable bits may have been changed */
#define POPCSRS \
	MOV (32*XLEN)(R2), R4; MOV R4, CSR(SSTATUS);\
	MOV (33*XLEN)(R2), R4; MOV R4, CSR(SIE);\
	MOV (34*XLEN)(R2), R4; MOV R4, CSR(SCAUSE);\
	MOV (35*XLEN)(R2), R4; MOV R4, CSR(STVAL);\
	MOV   0(R2), R4; MOV R4, CSR(SEPC)

/*
 * cope with possibly-nested supervisor-mode faults, from user or kernel mode.
 * the trap mechanism saved the return address in CSR(SEPC), and masked
 * interrupts.
 * SP should be valid on entry, but doesn't seem to always be!
 */
TEXT strap(SB), 1, $-4
	/* this should be a no-op if we trapped from kernel mode. */
	CSRRW	CSR(SSCRATCH), R(MACH), R(MACH)	/* save old m, load new m */

	/* stash a few regs in m->regsave to free working registers */
	MOV	R2, SAVESR2(R(MACH))		/* save SP */

	MOV	CSR(SSTATUS), R2
	AND	$Spp, R2
	BEQ	R2, fromuser		/* prev not super */

	/*
	 * trapped from kernel mode: R2, SB, MACH, and USER were all okay,
	 * in theory.  this can't be a system call.
	 * we shall use the current kernel stack.
	 */
	MOV	SAVESR2(R(MACH)), R2		/* restore kernel sp at entry */
	SUB	$UREGSIZE, R2			/* room for pc, regs, csrs */
	PUSHALLS

	MOV	$setSB(SB), R3	/* should be redundant, barring disaster */
/**/	CALLTRAP  /* C compiler is free to write regs but R2, SB, USER, MACH */
	CSRRC	CSR(SSTATUS), $Sie, R0;	/* back to splhi */

	/* pop regs except R2, SB, USER, MACH; pop CSRs */
	POPCSRS
	POPR8_31

	/*
	 * be sure to pop registers from Ureg, so that changes to this
	 * Ureg elsewhere (e.g. trap) are reflected.
	 * SP must point at a Ureg.
	 */
	/* MACH, SB are still correct; USER might not be after syscall? */
	MOV	(2*XLEN)(R(MACH)), R(USER)	/* up = m->proc */
	MOV	(1*XLEN)(R2), R1
	MOV	(5*XLEN)(R2), R5
	MOV	(4*XLEN)(R2), R4	/* restore saved R4 */
	MOV	(2*XLEN)(R2), R2
//	CSRRW	CSR(SSCRATCH), R(MACH), R(MACH)	/* revert (redundant) */
	SRET				/* to CSR(SEPC) address */

	/*
	 * trapped from user mode: SB, MACH, R2, USER all needed reloading.
	 * old MACH value is in CSR(SSCRATCH).  R2 is saved in regsave.
	 */
fromuser:
	/* stash a few regs in m->regsave to free working registers */
	MOV	R4, SAVESR4(R(MACH))	/* assembler temp */
	MOV	R6, SAVESR6(R(MACH))	/* save old R6 (now up) */

	MOV	(2*XLEN)(R(MACH)), R(USER) /*  up = m->proc; up is R6 */

	/*
	 * can't use user-mode stack.  switch to up->kstack.
	 */
	MOV	(2*XLEN)(R(USER)), R2	/* switch to empty up->kstack */
	ADD	$KSTACK, R2
	AND	$~7, R2			/* vlong alignment */

	SUB	$UREGSIZE, R2		/* room for pc, regs, csrs */
	PUSHALLS			/* patches R2 from regsave into Ureg */
	MOV	$setSB(SB), R3
	MOV	SAVESR4(R(MACH)), R4
	MOV	R4, (4*XLEN)(R2)	/* patch old R4 into Ureg */
	MOV	SAVESR6(R(MACH)), R9
	MOV	R9, (6*XLEN)(R2)	/* patch old R6 into Ureg */

	/* sched could be called and processes switched (e.g, syscall) */
/**/	CALLTRAP			/* C compiler is free to write regs */
	CSRRC	CSR(SSTATUS), $Sie, R0;	/* back to splhi */
syscallret:
	/* SP must point at our Ureg here.  up may have changed since entry. */
	/*
	 * before a syscall, the compiler pushed any registers it cared about
	 * (other than those it considers global, like SB, R2, MACH, USER),
	 * so we needn't save nor restore most of them before returning
	 * from a syscall.  We do need to propagate ureg->pc and ureg->sp
	 * back into user registers.
	 */
	POPCSRS
	POPR8_31
	/*
	 * be sure to pop user's registers from Ureg, so that changes to this
	 * Ureg elsewhere (e.g. sysexec) are reflected back to the user
	 * process.
	 */
	MOV	(1*XLEN)(R2), R1	/* restore user's link reg */
	MOV	(5*XLEN)(R2), R5
	MOV	(6*XLEN)(R2), R6
	MOV	(4*XLEN)(R2), R4	/* restore saved temporary */
	MOV	(3*XLEN)(R2), R3	/* restore user's SB */
	MOV	(2*XLEN)(R2), R2
	CSRRW	CSR(SSCRATCH), R(MACH), R(MACH) /* restore old R7, reload saved m */
	SRET				/* to CSR(SEPC) address */
	JMP	0(PC)

/* rfork return for child processes */
TEXT sysrforkret(SB), 1, $-4
	CSRRC	CSR(SSTATUS), $Sie, R0;	/* back to splhi */

	/* SP must point at our Ureg here.  up may have changed since entry. */
	MOV	R0, (ARG*XLEN)(R2)	/* zero saved R(ARG) in ureg */
	JMP	syscallret

/*
 * switch to user mode with stack pointer from sp+0(FP) at start of text.
 * used to start process 1 (init).
 */
TEXT touser(SB), 1, $-4
	/* ensure intrs see consistent kernel or user regs */
	MOV	$(Ube|Sie|Spp), R12	/* little endian, prev mode = user */
	CSRRC	CSR(SSTATUS), R12, R0

	FENCE
	FENCE_I

	MOV	$(Spie|Sum|Mxr), R12
	CSRRS	CSR(SSTATUS), R12, R0	/* new flags: allow user interrupts */
	MOV	R0, LINK
	MOV	$(UTZERO+0x20), R12
	MOV	R12, CSR(SEPC)		/* new pc */
	MOV	R(ARG), R2		/* new sp */
	SRET
