/*
 * microsemi/chip polarfire icicle dependencies
 */
#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"

#include "riscv.h"

/*
 * wait for *vp to change from val, or for an interrupt, on machines
 * that lack monitor & mwait instructions.
 * shouldn't be needed on hardware; VMs might need it.
 */
static int
portwaitfor(int *vp, int val)
{
	int i;
	Mpl pl;

	pl = splhi();
	if (sys->nonline <= 1)
		halt();
	else {
		/* is this cpu scheduling, thus done initialising? drop PL. */
		if (up)
			spllo();
		/*
		 * How many times round this loop?
		 */
		for(i = 10000; *vp == val && i > 0; i--)
			pause();
	}
	splx(pl);
	return *vp;
}

/* interface to the port code */
void
portmonitor(void *)
{
}

/*
 * only if mwait can be interrupted will we be able to
 * detect lock loops at high PL.
 *
 * this may seem to be a form of optimizing the idle loop, but mwait
 * should consume less power and fewer instructions on many-core systems
 * with more lock contention (e.g., NPROC=24 mk).
 */
void
portmwait(void)
{
}

int (*waitfor)(int*, int) = portwaitfor;
void (*mwaitforlock)(Lock *, ulong) = portmwaitforlock;

void
cpuidinit(void)
{
	if (m->machno == 0)
		print("%d-bit virtual addresses in use\n", Vaddrbits);
}

static int mwords[1];
static int *mwaitwd = mwords;			/* make safe from the start */

/*
 *  put the processor in the halt state if we've no processes to run.
 *  an interrupt or change to *mwaitwd will get us going again.
 *  called at spllo from proc.c, but also in infinite loops elsewhere.
 */
void
idlehands(void)
{
	waitfor(mwaitwd, *mwaitwd);
}

/*
 * wake any cpus in mwait state.  an interrupt will make one cpu break
 * out of mwait, but a qunlock probably should wake any idlers too.
 */
void
idlewake(void)
{
}

static ulong
prifexts(ulong isa, ulong exts, char *name)
{
	if ((isa & exts) == exts) {
		print(name);
		isa &= ~exts;
	}
	return isa;
}

#define EXT(let)	(1 << ((let) - 'A'))
#define EXTIMA		(EXT('I') | EXT('M') | EXT('A'))
#define EXTFD		(EXT('F') | EXT('D'))

uvlong cpuhz;

vlong
archhz(void)
{
	int bit;
	ulong isa;

	if(m->machno != 0)
		return sys->machptr[0]->cpuhz; /* cpus have to be ~identical */
	print("cpu%d: risc-v RV%d", m->machno, BI2BY*(int)sizeof(uintptr));
	if (machine_mode) {
		isa = prifexts(misa, EXTIMA|EXTFD, "G");
		isa = prifexts(isa, EXTIMA, "IMA");
		isa = prifexts(isa, EXTFD, "FD");
		isa = prifexts(isa, EXT('C'), "C");
		isa &= MASK('z'+1-'a');		/* isolate extensions */
		for (bit = 0; bit < 'z'+1-'a'; bit++)
			if (isa & bit)
				print("%c", 'A' + bit);
	} else
		print("G");			/* can't easily tell */
	print(" at %d MHz\n", (uint)(cpuhz / (1000*1000)));
	cpuidinit();
	return cpuhz;
}

static void
addpgsz(int lg2)
{
	int npg;

	assert(m->npgsz < NPGSZ);
	npg = m->npgsz++;
	m->pgszlg2[npg] = lg2;
	m->pgszmask[npg] = MASK(lg2);
}

int
archmmu(void)
{
	if (m->npgsz)
		return m->npgsz;
	CTASSERT(PGSZ == 4*KB, PGSZ_4K);
	addpgsz(PGSHFT);
	addpgsz(SUPPGSHFT);
	return m->npgsz;
}

static int
fmtP(Fmt* f)
{
	uintmem pa;

	pa = va_arg(f->args, uintmem);

	if(f->flags & FmtSharp)
		return fmtprint(f, "%#8.8lux", pa);

	return fmtprint(f, "%lud", pa);
}

static int
fmtL(Fmt* f)
{
	return fmtprint(f, "%#8.8lux", va_arg(f->args, Mpl));
}

static int
fmtR(Fmt* f)
{
	return fmtprint(f, "%#8.8lux", va_arg(f->args, Mreg));
}

void
archfmtinstall(void)
{
	/*
	 * Architecture-specific formatting. Not as neat as they
	 * could be (e.g. there's no defined type for a 'register':
	 *	L - Mpl, mach priority level
	 *	P - uintmem, physical address
	 *	R - register
	 * With a little effort these routines could be written
	 * in a fairly architecturally-independent manner, relying
	 * on the compiler to optimise-away impossible conditions,
	 * and/or by exploiting the innards of the fmt library.
	 */
	fmtinstall('P', fmtP);

	fmtinstall('L', fmtL);
	fmtinstall('R', fmtR);
}

/* delay for at least microsecs, avoiding the watchdog if necessary */
void
microdelay(vlong microsecs)
{
	uvlong t, now, nxtdog, mhz;

	if (microsecs <= 0)
		return;
	now = rdtsc();
	assert(m);
	mhz = m->cpumhz;
	if (mhz == 0)
		mhz = 2000;
	t = now + mhz * microsecs;
	/* when islo(), clock interrupts will restart the dog */
	if (watchdogon && watchdog && m->machno == 0 && !islo()) {
		nxtdog = 0;
		for (now = rdtsc(); now < t; now = rdtsc()) {
			if (now > nxtdog) {
				watchdog->restart();
				nxtdog = now + Wdogms*1000ll*mhz;
			}
			pause();
		}
	} else
		while (rdtsc() < t)
			pause();
}

void
millidelay(int millisecs)
{
	microdelay(1000ll * millisecs);
}


int
clz(Clzuint n)			/* count leading zeroes */
{
	int cnt;
	uint hibits;

	if (n == 0)
		return Clzbits;
	cnt = 0;
	for (hibits = Clzbits/2; hibits > 0; hibits /= 2)
		if ((n & (VMASK(hibits) << (Clzbits - hibits))) == 0) {
			/* high bits are zero; count and toss them */
			n <<= hibits;
			cnt += hibits;
		}
	return cnt;
}
